# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Yan Jia)

import logging

logger = logging.getLogger(__name__)


class TokenDict:
    def __init__(self, dict_path, unk=""):
        assert dict_path != ""
        self.id2word, self.word2id = self.read_dict(dict_path)
        self.unk = unk
        assert unk == "" or unk in self.word2id
        self.unkid = self.word2id[unk] if unk else -1

    def get(self, key, default):
        if type(default) == str:
            default = self.word2id[default]
        return self.word2id.get(key, default)

    def __getitem__(self, key):
        if type(key) == str:
            if self.unk:
                return self.word2id.get(key, self.word2id[self.unk])
            else:
                return self.word2id[key]
        elif type(key) == int:
            return self.id2word[key]
        else:
            raise TypeError("Key should be str or int")

    def __len__(self):
        return len(self.id2word)

    def __contains__(self, query):
        if type(query) == str:
            return query in self.word2id
        elif type(query) == int:
            return query in self.id2word
        else:
            raise TypeError("query should be str or int")

    def read_dict(self, dict_path):
        id2word, word2id = [], {}
        with open(dict_path, encoding="utf8") as f:
            for i, line in enumerate(f):
                tokens = line.strip().split()
                if len(tokens) >= 2:
                    word, index = tokens[0], int(tokens[1])
                elif len(tokens) == 1:
                    word, index = tokens[0], i
                else:
                    logger.info(
                        f"Find empty line or space '{line.strip()}'"
                        f" in {dict_path}:L{i}, set to ' '"
                    )
                    word, index = " ", i
                assert len(id2word) == index
                assert len(word2id) == index
                if word == "<space>":
                    logger.info(
                        f"NOTE: Find <space> in {dict_path}:L{i} and convert it to ' '"
                    )
                    word = " "
                word2id[word] = index
                id2word.append(word)
        assert len(id2word) == len(word2id)
        return id2word, word2id



from .audio import (
    Audio,
    AudioChannel,
    AudioChannelList,
    AudioChunk,
    AudioChunkList,
    AudioList,
    AudioSpan,
    AudioSpanList,
    AudioToken,
    AudioTokenList,
)
from .eval import AudioEvalExample, AudioEvalResult
from .stream import AudioBytesStream
from .waveform import Waveform

__all__ = [
    "Waveform",
    "Audio",
    "AudioList",
    "AudioChannel",
    "AudioChannelList",
    "AudioSpan",
    "AudioSpanList",
    "AudioToken",
    "AudioTokenList",
    "AudioChunk",
    "AudioChunkList",
    "AudioEvalResult",
    "AudioEvalExample",
    "AudioBytesStream",
]



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Kai Huang)

from collections import deque
from dataclasses import dataclass
import enum


@dataclass
class StreamVadFrameResult:
    frame_idx: int  # 1-based
    is_speech: bool
    raw_prob: float
    smoothed_prob: float
    is_speech_start: bool = False
    is_speech_end: bool = False
    speech_start_frame: int = -1  # 1-based
    speech_end_frame: int = -1  # 1-based


@enum.unique
class VadState(enum.Enum):
    SILENCE = 0
    POSSIBLE_SPEECH = 1
    SPEECH = 2
    POSSIBLE_SILENCE = 3


class StreamVadPostprocessor:
    def __init__(
        self,
        smooth_window_size,
        speech_threshold,
        pad_start_frame,
        min_speech_frame,
        max_speech_frame,
        min_silence_frame,
    ):
        self.smooth_window_size = max(1, smooth_window_size)
        self.speech_threshold = speech_threshold
        self.pad_start_frame = max(self.smooth_window_size, pad_start_frame)
        self.min_speech_frame = min_speech_frame
        self.max_speech_frame = max_speech_frame
        self.min_silence_frame = min_silence_frame
        self.reset()

    def reset(self):
        self.frame_cnt = 0
        # smooth window
        self.smooth_window = deque()
        self.smooth_window_sum = 0.0
        # state transition
        self.state = VadState.SILENCE
        self.speech_cnt = 0
        self.silence_cnt = 0
        self.hit_max_speech = False
        self.last_speech_start_frame = -1
        self.last_speech_end_frame = -1

    def process_one_frame(self, raw_prob):
        assert type(raw_prob) == float
        assert 0.0 <= raw_prob and raw_prob <= 1.0
        self.frame_cnt += 1

        smoothed_prob = self.smooth_prob(raw_prob)

        is_speech = self.apply_threshold(smoothed_prob)

        result = StreamVadFrameResult(
            frame_idx=self.frame_cnt,
            is_speech=is_speech,
            raw_prob=round(raw_prob, 3),
            smoothed_prob=round(smoothed_prob, 3),
        )

        result = self.state_transition(is_speech, result)

        return result

    def smooth_prob(self, prob):
        if self.smooth_window_size <= 1:
            return prob
        self.smooth_window.append(prob)
        self.smooth_window_sum += prob
        if len(self.smooth_window) > self.smooth_window_size:
            left = self.smooth_window.popleft()
            self.smooth_window_sum -= left
        smoothed_prob = self.smooth_window_sum / len(self.smooth_window)
        return smoothed_prob

    def apply_threshold(self, prob):
        return int(prob >= self.speech_threshold)

    def state_transition(self, is_speech, result):
        if self.hit_max_speech:
            result.is_speech_start = True
            result.speech_start_frame = self.frame_cnt
            self.last_speech_start_frame = result.speech_start_frame
            self.hit_max_speech = False

        if self.state == VadState.SILENCE:
            if is_speech:
                self.state = VadState.POSSIBLE_SPEECH
                self.speech_cnt += 1
            else:
                self.silence_cnt += 1
                self.speech_cnt = 0

        elif self.state == VadState.POSSIBLE_SPEECH:
            if is_speech:
                self.speech_cnt += 1
                if self.speech_cnt >= self.min_speech_frame:
                    self.state = VadState.SPEECH
                    result.is_speech_start = True
                    result.speech_start_frame = max(
                        1,
                        self.frame_cnt - self.speech_cnt + 1 - self.pad_start_frame,
                        self.last_speech_end_frame + 1,
                    )
                    self.last_speech_start_frame = result.speech_start_frame
                    self.silence_cnt = 0
            else:
                self.state = VadState.SILENCE
                self.silence_cnt = 1
                self.speech_cnt = 0

        elif self.state == VadState.SPEECH:
            self.speech_cnt += 1
            if is_speech:
                self.silence_cnt = 0
                if self.speech_cnt >= self.max_speech_frame:
                    self.hit_max_speech = True
                    self.speech_cnt = 0
                    result.is_speech_end = True
                    result.speech_end_frame = self.frame_cnt
                    result.speech_start_frame = self.last_speech_start_frame
                    self.last_speech_start_frame = -1
                    self.last_speech_end_frame = result.speech_end_frame
            else:
                self.state = VadState.POSSIBLE_SILENCE
                self.silence_cnt += 1

        elif self.state == VadState.POSSIBLE_SILENCE:
            self.speech_cnt += 1
            if is_speech:
                self.state = VadState.SPEECH
                self.silence_cnt = 0
                if self.speech_cnt >= self.max_speech_frame:
                    self.hit_max_speech = True
                    self.speech_cnt = 0
                    result.is_speech_end = True
                    result.speech_end_frame = self.frame_cnt
                    result.speech_start_frame = self.last_speech_start_frame
                    self.last_speech_start_frame = -1
                    self.last_speech_end_frame = result.speech_end_frame

            else:
                self.silence_cnt += 1
                if self.silence_cnt >= self.min_silence_frame:
                    self.state = VadState.SILENCE
                    result.is_speech_end = True
                    result.speech_end_frame = self.frame_cnt
                    result.speech_start_frame = self.last_speech_start_frame
                    self.last_speech_end_frame = result.speech_end_frame
                    self.last_speech_start_frame = -1
                    self.speech_cnt = 0

        return result



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Yan Jia)

import glob
import logging
import os

logger = logging.getLogger(__name__)


def get_wav_info(args):
    """
    Returns:
        wavs: list of (uttid, wav_path)
    """

    def base(p):
        return os.path.basename(p).replace(".wav", "")

    if args.wav_path:
        wavs = [(base(args.wav_path), args.wav_path)]
    elif args.wav_paths and len(args.wav_paths) >= 1:
        wavs = [(base(p), p) for p in sorted(args.wav_paths)]
    elif args.wav_scp:
        wavs = [line.strip().split() for line in open(args.wav_scp)]
        if args.sort_wav_by_dur:
            logger.info("Sort wav by duration...")
            utt2dur = os.path.join(os.path.dirname(args.wav_scp), "utt2dur")
            if os.path.exists(utt2dur):
                utt2dur = [l.strip().split() for l in open(utt2dur)]
                utt2dur = {l[0]: float(l[1]) for l in utt2dur if len(l) == 2}
                wavs = sorted(wavs, key=lambda x: -utt2dur[x[0]])
                logger.info("Sort Done")
            else:
                logger.info(f"Not find {utt2dur}, un-sort")
    elif args.wav_dir:
        wavs = glob.glob(f"{args.wav_dir}/**/*.wav", recursive=True)
        wavs = [(base(p), p) for p in sorted(wavs)]
    else:
        raise ValueError("Please provide valid wav info")
    logger.info(f"#wavs={len(wavs)}")
    return wavs



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence


class LstmLm(nn.Module):
    @classmethod
    def from_args(cls, args):
        args.padding_idx = 2
        args.sos_id = 3
        args.eos_id = 4
        return cls(args)

    def __init__(self, args):
        super().__init__()
        self.embedding = nn.Embedding(
            args.idim, args.embedding_dim, padding_idx=args.padding_idx
        )
        self.lstm = nn.LSTM(
            args.embedding_dim,
            args.hidden_size,
            args.num_layers,
            batch_first=True,
            dropout=args.dropout,
        )
        self.fc_in_dim = args.embedding_dim
        self.fc = nn.Linear(args.embedding_dim, args.odim)

        self._tie_weights(args)
        self.sos_id = args.sos_id
        self.eos_id = args.eos_id
        self.ignore_index = args.padding_idx

    @torch.jit.ignore
    def _tie_weights(self, args):
        if args.tie_weights:
            if self.fc_in_dim != args.embedding_dim or args.idim != args.odim:
                raise ValueError(
                    "When using the tied flag, fc_in_dim must be equal to embedding_dim"
                )
            self.fc.weight = self.embedding.weight

    @torch.jit.export
    def init_hidden(self, tensor, batch_size):
        # type: (Tensor, int) -> Tuple[Tensor, Tensor]
        return (
            tensor.new_zeros(
                self.lstm.num_layers, batch_size, self.lstm.hidden_size
            ).float(),
            tensor.new_zeros(
                self.lstm.num_layers, batch_size, self.lstm.hidden_size
            ).float(),
        )

    @torch.jit.export
    def forward_model(self, padded_inputs, lengths=None, hidden=None):
        # type: (Tensor, Optional[Tensor], Optional[Tuple[Tensor, Tensor]]) -> Tuple[Tensor, Tuple[Tensor, Tensor]]
        # Embedding Layer
        padded_inputs = self.embedding(padded_inputs)  # N, T, D
        # LSTM Layers
        if lengths is None:
            output, new_hidden = self.lstm(padded_inputs, hidden)
        else:
            lengths = lengths.cpu().int()
            total_length = padded_inputs.size(1)  # get the max sequence length
            packed_input = pack_padded_sequence(
                padded_inputs, lengths, batch_first=True, enforce_sorted=False
            )
            # self.lstm.flatten_parameters()
            packed_output, new_hidden = self.lstm(packed_input, hidden)
            output, _ = pad_packed_sequence(
                packed_output, batch_first=True, total_length=total_length
            )
        # Output Layer
        score = self.fc(output)  # (N, T, V)
        return score, new_hidden



from __future__ import annotations

from pathlib import Path
from typing import Any, List

import numpy as np
from pydantic import ConfigDict, Field
import torch
from typing_extensions import Self

from fasr.config import registry
from fasr.data import AudioToken, AudioTokenList
from fasr.model import ASRModel


def _resolve_dtype(dtype: str | torch.dtype | None) -> torch.dtype | None:
    if dtype is None or isinstance(dtype, torch.dtype):
        return dtype
    value = dtype.strip()
    mapping = {
        "float16": torch.float16,
        "fp16": torch.float16,
        "bfloat16": torch.bfloat16,
        "bf16": torch.bfloat16,
        "float32": torch.float32,
        "fp32": torch.float32,
    }
    if value not in mapping:
        raise ValueError(f"Unsupported dtype: {dtype!r}")
    return mapping[value]


def _result_to_text(item: Any) -> str:
    if isinstance(item, str):
        return item.strip()
    if isinstance(item, dict):
        text = item.get("text")
        return "" if text is None else str(text).strip()
    text_attr = getattr(item, "text", None)
    if text_attr is not None:
        return str(text_attr).strip()
    return str(item).strip()


class _Qwen3ForASRBase(ASRModel):
    """Qwen3 ASR 基类（无时间戳）。"""

    endpoint: str = "modelscope"

    qwen3: Any | None = Field(default=None, exclude=True)
    max_new_tokens: int = 4096
    max_inference_batch_size: int = 128
    gpu_memory_utilization: float = 0.8

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_assignment=True)

    def transcribe(
        self,
        batch: List[np.ndarray | torch.Tensor],
        **kwargs,
    ) -> List[AudioTokenList]:
        if self.qwen3 is None:
            raise RuntimeError("Call from_checkpoint before transcribe")
        sample_rate = int(kwargs.get("sample_rate", 16000))
        out: list[AudioTokenList] = []
        for waveform in batch:
            if isinstance(waveform, torch.Tensor):
                wav = waveform.detach().cpu().float().numpy()
            else:
                wav = np.asarray(waveform)
            wav = wav.squeeze()
            result = self.qwen3.transcribe((wav, sample_rate))
            if isinstance(result, list):
                text = _result_to_text(result[0]) if result else ""
            else:
                text = _result_to_text(result)
            # Qwen3 当前实现不提供时间戳，整段文本作为一个 token 输出
            out.append(AudioTokenList([AudioToken(text=text)]))
        return out

    def from_checkpoint(
        self,
        checkpoint_dir: str | Path | None = None,
        device_map: str | None = "auto",
        dtype: str | torch.dtype | None = "bfloat16",
        gpu_memory_utilization: float = 0.8,
        max_inference_batch_size: int = 128,
        max_new_tokens: int = 4096,
        max_model_len: int | None = None,
        **kwargs,
    ) -> Self:
        try:
            from qwen_asr import Qwen3ASRModel
        except ImportError as exc:
            raise ImportError(
                "Please install qwen-asr first: pip install qwen-asr[vllm]"
            ) from exc

        if not checkpoint_dir:
            checkpoint_dir = self.download_checkpoint()
        checkpoint_dir = Path(checkpoint_dir)
        if not checkpoint_dir.exists():
            raise FileNotFoundError(f"Checkpoint not found: {checkpoint_dir}")
        self.max_inference_batch_size = max_inference_batch_size
        self.max_new_tokens = max_new_tokens
        self.qwen3 = Qwen3ASRModel.LLM(
            model=str(checkpoint_dir),
            gpu_memory_utilization=gpu_memory_utilization,
            max_inference_batch_size=max_inference_batch_size,  # Batch size limit for inference. -1 means unlimited. Smaller values can help avoid OOM.
            max_new_tokens=max_new_tokens,  # Maximum number of tokens to generate. Set a larger value for long audio input.
            max_model_len=max_model_len or max_new_tokens * 2,
        )
        return self

    def get_config(self):
        raise NotImplementedError

    def load(self, save_dir, **kwargs):
        raise NotImplementedError

    def save(self, save_dir, **kwargs):
        raise NotImplementedError


@registry.asr_models.register("qwen3_0_6b")
class Qwen3_06BForASR(_Qwen3ForASRBase):
    checkpoint: str = "Qwen/Qwen3-ASR-0.6B"


@registry.asr_models.register("qwen3_1_7b")
class Qwen3_17BForASR(_Qwen3ForASRBase):
    checkpoint: str = "Qwen/Qwen3-ASR-1.7B"



from __future__ import annotations

from pathlib import Path
from typing import Any, List

from pydantic import ConfigDict, Field
from typing_extensions import Self

from fasr.config import registry
from fasr.data import AudioSpan, AudioSpanList, Waveform
from fasr.model import VADModel


@registry.vad_models.register("firered")
class FireRedForVAD(VADModel):
    """FireRedVAD 非流式语音活动检测。

    权重目录需含 ``cmvn.ark`` 和 ``model.pth.tar``，
    与 ``FireRedVad.from_pretrained(model_dir)`` 一致。
    """

    checkpoint: str = "FireRedTeam/FireRedVAD"
    endpoint: str = "modelscope"

    firered: Any | None = Field(default=None, exclude=True)
    use_gpu: bool = False
    speech_threshold: float = 0.4

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_assignment=True)

    def detect(self, waveform: Waveform) -> List[AudioSpan]:
        wav = waveform.resample(16000) if waveform.sample_rate != 16000 else waveform
        pcm = wav.to_int16_pcm()
        audio_input = (pcm, 16000)
        result, _ = self.firered.detect(audio_input)
        segments = AudioSpanList()
        for start_s, end_s in result["timestamps"]:
            start_ms = int(round(start_s * 1000))
            end_ms = int(round(end_s * 1000))
            if end_ms > start_ms:
                seg = AudioSpan(start_ms=start_ms, end_ms=end_ms)
                seg.waveform = waveform.select_by_ms(start=start_ms, end=end_ms)
                segments.append(seg)
        return segments

    def from_checkpoint(
        self,
        checkpoint_dir: str | Path | None = None,
        use_gpu: bool = False,
        speech_threshold: float = 0.4,
        **kwargs,
    ) -> Self:
        from fireredvad.vad import FireRedVad, FireRedVadConfig

        _ = kwargs
        if not checkpoint_dir:
            checkpoint_dir = self.download_checkpoint()
        checkpoint_dir = Path(checkpoint_dir)
        vad_dir = checkpoint_dir / "VAD"
        if vad_dir.is_dir():
            checkpoint_dir = vad_dir
        if not checkpoint_dir.exists():
            raise FileNotFoundError(f"Checkpoint not found: {checkpoint_dir}")
        config = FireRedVadConfig(
            use_gpu=use_gpu,
            speech_threshold=speech_threshold,
        )
        self.firered = FireRedVad.from_pretrained(str(checkpoint_dir), config=config)
        self.use_gpu = use_gpu
        self.speech_threshold = speech_threshold
        return self

    def get_config(self):
        raise NotImplementedError

    def load(self, save_dir, **kwargs):
        raise NotImplementedError

    def save(self, save_dir, **kwargs):
        raise NotImplementedError



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

import logging

import torch

logger = logging.getLogger(__name__)


def count_model_parameters(model):
    if not isinstance(model, torch.nn.Module):
        return 0, 0
    name = f"{model.__class__.__name__} {model.__class__}"
    num = sum(p.numel() for p in model.parameters() if p.requires_grad)
    size = num * 4.0 / 1024.0 / 1024.0  # float32, MB
    logger.info(f"#param of {name} is {num} = {size:.1f} MB (float32)")
    return num, size



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Kai Huang)

from dataclasses import dataclass
import logging
import os

import torch

from .core.audio_feat import AudioFeat
from .core.detect_model import DetectModel
from .core.vad_postprocessor import VadPostprocessor

logger = logging.getLogger(__name__)


@dataclass
class FireRedVadConfig:
    use_gpu: bool = False
    smooth_window_size: int = 5
    speech_threshold: float = 0.4
    min_speech_frame: int = 20
    max_speech_frame: int = 2000  # 20s
    min_silence_frame: int = 20
    merge_silence_frame: int = 0
    extend_speech_frame: int = 0
    chunk_max_frame: int = 30000  # 300s

    def __post_init__(self):
        if self.speech_threshold < 0 or self.speech_threshold > 1:
            raise ValueError("speech_threshold must be in [0, 1]")
        if self.min_speech_frame <= 0:
            raise ValueError("min_speech_frame must be positive")


class FireRedVad:
    @classmethod
    def from_pretrained(cls, model_dir, config=FireRedVadConfig()):
        # Build Feat Extractor
        cmvn_path = os.path.join(model_dir, "cmvn.ark")
        audio_feat = AudioFeat(cmvn_path)

        # Build Model
        vad_model = DetectModel.from_pretrained(model_dir)
        if config.use_gpu:
            vad_model.cuda()
        else:
            vad_model.cpu()

        # Build Postprocessor
        vad_postprocessor = VadPostprocessor(
            config.smooth_window_size,
            config.speech_threshold,
            config.min_speech_frame,
            config.max_speech_frame,
            config.min_silence_frame,
            config.merge_silence_frame,
            config.extend_speech_frame,
        )
        return cls(audio_feat, vad_model, vad_postprocessor, config)

    def __init__(self, audio_feat, vad_model, vad_postprocessor, config):
        self.audio_feat = audio_feat
        self.vad_model = vad_model
        self.vad_postprocessor = vad_postprocessor
        self.config = config

    def detect(self, audio, do_postprocess=True):
        # Extract feat
        feats, dur = self.audio_feat.extract(audio)
        if self.config.use_gpu:
            feats = feats.cuda()

        # Model inference
        if feats.size(0) <= self.config.chunk_max_frame:
            probs, _ = self.vad_model.forward(feats.unsqueeze(0))
            probs = probs.cpu().squeeze()  # (T,)
        else:
            logger.warning(
                f"Too long input, split every {self.config.chunk_max_frame} frames"
            )
            chunk_probs = []
            chunks = feats.split(self.config.chunk_max_frame, dim=0)
            for chunk in chunks:
                chunk_prob, _ = self.vad_model.forward(chunk.unsqueeze(0))
                chunk_probs.append(chunk_prob.cpu())
            probs = torch.cat(chunk_probs, dim=1)
            probs = probs.squeeze()  # (T,)

        if not do_postprocess:
            return None, probs

        # Prob Postprocess
        decisions = self.vad_postprocessor.process(probs.tolist())
        starts_ends_s = self.vad_postprocessor.decision_to_segment(decisions, dur)

        # Format result
        result = {"dur": round(dur, 3), "timestamps": starts_ends_s}
        if isinstance(audio, str):
            result["wav_path"] = audio
        return result, probs



from pathlib import Path
from typing import Dict, Iterable

from funasr import AutoModel
from funasr_onnx.paraformer_online_bin import Paraformer as ParaformerONNX
from loguru import logger
import numpy as np
import torch
from typing_extensions import Self

from fasr.config import registry
from fasr.data import AudioChunk, AudioToken
from fasr.model import StreamASRModel


@registry.stream_asr_models.register("stream_paraformer")
@registry.stream_asr_models.register("stream_paraformer.torch")
class ParaformerForStreamASR(StreamASRModel):
    """Paraformer流式语音识别模型"""

    checkpoint: str = (
        "iic/speech_paraformer-large_asr_nat-zh-cn-16k-common-vocab8404-online"
    )
    endpoint: str = "modelscope"

    chunk_size_ms: int = 600  # chunk size in ms
    encoder_chunk_look_back: int = (
        4  # number of chunks to lookback for encoder self-attention
    )
    decoder_chunk_look_back: int = 1
    sample_rate: int = 16000

    paraformer: AutoModel | None = None

    def transcribe_chunk(self, chunk: AudioChunk) -> Iterable[AudioToken]:
        if chunk.is_start and chunk.stream_id in self.states:
            logger.warning(
                "found start chunk with existing state, this is not expected"
            )
            del self.states[chunk.stream_id]
        state: Dict = self.states.get(chunk.stream_id, {})
        waveform = chunk.waveform
        if waveform.sample_rate != self.sample_rate:
            waveform = waveform.resample(self.sample_rate)
        sample_rate = waveform.sample_rate
        data = waveform.data
        chunk_size = int(self.chunk_size_ms * sample_rate / 1000)
        buffer = state.get("buffer", np.array([], dtype=np.float32))
        buffer = np.concatenate([buffer, data], axis=0)
        cache = state.get("cache", {})
        if chunk.is_last:
            stream = self.paraformer.generate(
                input=buffer,
                cache=cache,
                is_final=True,
                chunk_size=[0, 10, 5],
                encoder_chunk_look_back=self.encoder_chunk_look_back,
                decoder_chunk_look_back=self.decoder_chunk_look_back,
            )
            for result in stream:
                if result["text"]:
                    yield AudioToken(text=result["text"])
        else:
            while len(buffer) > chunk_size:
                input_data = buffer[:chunk_size]
                buffer = buffer[chunk_size:]
                stream = self.paraformer.generate(
                    input=input_data,
                    cache=cache,
                    is_final=False,
                    chunk_size=[0, 10, 5],  # chunk size 10 * 60ms
                    encoder_chunk_look_back=self.encoder_chunk_look_back,
                    decoder_chunk_look_back=self.decoder_chunk_look_back,
                )
                for result in stream:
                    if result["text"]:
                        yield AudioToken(text=result["text"])
        state["buffer"] = buffer
        state["cache"] = cache
        self.states[chunk.stream_id] = state
        if chunk.is_last:
            self.states.pop(chunk.stream_id)

    def reset(self):
        self.states.clear()

    def from_checkpoint(
        self,
        checkpoint_dir: str | Path | None = None,
        disable_update: bool = True,
        disable_log: bool = True,
        disable_pbar: bool = True,
        compile: bool = False,
        **kwargs,
    ) -> Self:
        if not checkpoint_dir:
            checkpoint_dir = self.download_checkpoint()
        checkpoint_dir = Path(checkpoint_dir)
        if not checkpoint_dir.exists():
            raise FileNotFoundError(f"Checkpoint not found: {checkpoint_dir}")
        self.paraformer = AutoModel(
            model=checkpoint_dir,
            disable_update=disable_update,
            disable_log=disable_log,
            disable_pbar=disable_pbar,
            **kwargs,
        )
        if compile:
            logger.info("using torch.compile for stream paraformer")
            self.paraformer.model = torch.compile(
                self.paraformer.model, fullgraph=True, dynamic=True
            )
        return self


@registry.stream_asr_models.register("stream_paraformer.onnx")
class ParaformerForStreamASROnnx(StreamASRModel):
    """Paraformer流式语音识别模型"""

    checkpoint: str = (
        "iic/speech_paraformer-large_asr_nat-zh-cn-16k-common-vocab8404-online"
    )
    endpoint: str = "modelscope"
    chunk_size_ms: int = 600  # chunk size in ms
    encoder_chunk_look_back: int = (
        4  # number of chunks to lookback for encoder self-attention
    )
    decoder_chunk_look_back: int = 1
    sample_rate: int = 16000
    paraformer: ParaformerONNX = None

    def transcribe_chunk(self, chunk: AudioChunk) -> Iterable[AudioToken]:
        if chunk.is_start and chunk.stream_id in self.states:
            logger.warning(
                "found start chunk with existing state, this is not expected"
            )
            del self.states[chunk.stream_id]
        state = self.states.get(chunk.stream_id, {})
        waveform = chunk.waveform
        if waveform.sample_rate != self.sample_rate:
            waveform = waveform.resample(self.sample_rate)
        sample_rate = waveform.sample_rate
        data = waveform.data
        chunk_size = int(self.chunk_size_ms * sample_rate / 1000)
        buffer = state.get("buffer", np.array([], dtype=np.float32))
        buffer = np.concatenate([buffer, data], axis=0)
        cache = state.get("cache", {})
        param_dict = {"cache": cache}
        if chunk.is_last:
            param_dict["is_final"] = True
            stream = self.paraformer(
                audio_in=buffer,
                param_dict=param_dict,
            )
            for result in stream:
                if result["preds"]:
                    for t in result["preds"][1]:
                        yield AudioToken(text=t)
        else:
            while len(buffer) > chunk_size:
                audio_in = buffer[:chunk_size]
                buffer = buffer[chunk_size:]
                stream = self.paraformer(
                    audio_in=audio_in,
                    param_dict=param_dict,
                )
                for result in stream:
                    if result["preds"]:
                        for t in result["preds"][1]:
                            yield AudioToken(text=t)
        state["buffer"] = buffer
        state["cache"] = cache
        self.states[chunk.stream_id] = state
        if chunk.is_last:
            self.states.pop(chunk.stream_id)

    def from_checkpoint(
        self,
        checkpoint_dir: str | Path | None = None,
        quantize: bool = False,
        device: str | None = None,
        intra_op_num_threads: int = 4,
        **kwargs,
    ) -> Self:
        if not checkpoint_dir:
            checkpoint_dir = self.download_checkpoint()
        checkpoint_dir = Path(checkpoint_dir)
        if not checkpoint_dir.exists():
            raise FileNotFoundError(f"Checkpoint not found: {checkpoint_dir}")
        if not device:
            device = "cuda" if torch.cuda.is_available() else "cpu"
        if device == "cuda":
            device_id = 0
        else:
            device_id = -1
        self.paraformer = ParaformerONNX(
            model_dir=checkpoint_dir,
            chunk_size=[0, 10, 5],
            quantize=quantize,
            device_id=device_id,
            intra_op_num_threads=intra_op_num_threads,
            **kwargs,
        )
        return self



from abc import abstractmethod

import numpy as np
from pydantic import ConfigDict

from fasr.utils.base import CheckpointMixin, IOMixin


class BaseWaveformPreprocessor(IOMixin, CheckpointMixin):
    model_config: ConfigDict = ConfigDict(
        arbitrary_types_allowed=True, validate_assignment=True
    )

    @abstractmethod
    def process_waveform(self, waveform: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def from_checkpoint(self, checkpoint_dir: str, **kwargs):
        raise NotImplementedError

    def get_config(self):
        raise NotImplementedError

    def save(self, save_dir):
        raise NotImplementedError

    def load(self, save_dir):
        raise NotImplementedError



from __future__ import annotations

from pathlib import Path
from typing import Any, Generator, Literal

from docarray import BaseDoc, DocList
from docarray.utils.filter import filter_docs
from pydantic import Field, model_validator
from typing_extensions import Self

from .audio import Audio, AudioChannel, AudioList


class AudioEvalExample(BaseDoc):
    """Audio example object that represents an example of input and target audio."""

    x: Audio = Field(..., description="Input audio.")
    y: Audio = Field(..., description="Target audio.")
    is_badcase: bool = Field(False, description="whether the example is a badcase")

    @model_validator(mode="after")
    def validate_audio_example(self):
        if len(self.x.channels) != len(self.y.channels):
            raise ValueError(
                "Input and target audio must have the same number of channels."
            )
        return self

    def char_error_rate(
        self,
        ignore_space: bool = True,
        ignore_punc: bool = True,
        normalize: bool = False,
    ) -> float:
        """Compute the character error rate (CER) between the input and target audio on all the channels.

        Args:
            ignore_space (bool, optional): Whether to ignore space. Defaults to True.
            ignore_punc (bool, optional): Whether to ignore punctuation. Defaults
            normalize (bool, optional): Whether to normalize the text. Defaults to False.
        Returns:
            float: Character error rate.

        """
        from fasr.utils.metric import compute_char_error_rate

        cer = 0
        for i in range(len(self.x.channels)):
            x = self.x.channels[i].text
            y = self.y.channels[i].text
            err = compute_char_error_rate(
                x=x,
                y=y,
                ignore_space=ignore_space,
                ignore_punc=ignore_punc,
                normalize=normalize,
            )
            if err > 0:
                self.is_badcase = True
            cer += err

        return round(cer, 4)

    def clear_waveform(self) -> Self:
        """Clear the waveform of the input and target audio."""
        self.x.clear_waveform()
        self.y.clear_waveform()
        return self

    def display(self) -> None:
        """Display character-level differences between x and y for each channel.

        Uses edit distance alignment to highlight:
        - Green: correct (matching characters)
        - Red: substitution (different characters)
        - Yellow: insertion (extra characters in x)
        - Cyan: deletion (missing characters from y)
        """
        from fasr.utils.metric import match_token

        RED = "\033[91m"
        GREEN = "\033[92m"
        YELLOW = "\033[93m"
        CYAN = "\033[96m"
        RESET = "\033[0m"
        BOLD = "\033[1m"

        for i in range(len(self.x.channels)):
            x_text = self.x.channels[i].text or ""
            y_text = self.y.channels[i].text or ""

            print(f"{BOLD}Channel {i}{RESET}")

            if x_text == y_text:
                print(f"  x: {GREEN}{x_text}{RESET}")
                print(f"  y: {GREEN}{y_text}{RESET}")
                print()
                continue

            x_chars = list(x_text)
            y_chars = list(y_text)
            matched = match_token(x_chars, y_chars)

            matched_y = {m[0] for m in matched}
            matched_x = {m[1] for m in matched}

            x_display = ""
            for idx, ch in enumerate(x_chars):
                if idx in matched_x:
                    x_display += f"{GREEN}{ch}{RESET}"
                else:
                    x_display += f"{YELLOW}{ch}{RESET}"

            y_display = ""
            for idx, ch in enumerate(y_chars):
                if idx in matched_y:
                    y_display += f"{GREEN}{ch}{RESET}"
                else:
                    y_display += f"{CYAN}{ch}{RESET}"

            subs = []
            for ref_idx, hyp_idx in matched:
                if y_chars[ref_idx] != x_chars[hyp_idx]:
                    subs.append(
                        f"{RED}{y_chars[ref_idx]}{RESET} → {RED}{x_chars[hyp_idx]}{RESET}"
                    )

            ins_chars = [
                x_chars[idx] for idx in range(len(x_chars)) if idx not in matched_x
            ]
            del_chars = [
                y_chars[idx] for idx in range(len(y_chars)) if idx not in matched_y
            ]

            print(f"  x: {x_display}")
            print(f"  y: {y_display}")
            if subs:
                print(f"  {RED}替换: {', '.join(subs)}{RESET}")
            if ins_chars:
                print(f"  {YELLOW}插入(x多余): {''.join(ins_chars)}{RESET}")
            if del_chars:
                print(f"  {CYAN}删除(y缺失): {''.join(del_chars)}{RESET}")
            print()


class AudioEvalResult(DocList):
    @property
    def x(self) -> AudioList:
        return AudioList(docs=[example.x for example in self])

    @property
    def y(self) -> AudioList:
        return AudioList(docs=[example.y for example in self])

    def add_example(self, x: Audio | AudioList, y: Audio | AudioList) -> Self:
        """Add an audio example to the evaluation."""
        if isinstance(x, Audio):
            x = AudioList([x])
        if isinstance(y, Audio):
            y = AudioList([y])
        if len(x) != len(y):
            raise ValueError(
                "Input and target audio must have the same number of channels."
            )
        for x_audio, y_audio in zip(x, y):
            self.append(AudioEvalExample(x=x_audio, y=y_audio))
        return self

    def char_error_rate(
        self,
        ignore_space: bool = True,
        ignore_punc: bool = True,
        normalize: bool = False,
    ) -> float:
        """Compute the character error rate (CER) between the input and target audio on all the channels.
        Args:
            ignore_space (bool, optional): Whether to ignore space. Defaults to True.
            ignore_punc (bool, optional): Whether to ignore punctuation. Defaults to True.
            normalize (bool, optional): Whether to normalize the text. Defaults to False.
        Returns:
            float: Character error rate.
        """

        cer = 0
        for audio_example in self:
            audio_example: AudioEvalExample
            cer += audio_example.char_error_rate(
                ignore_space=ignore_space,
                ignore_punc=ignore_punc,
                normalize=normalize,
            )

        return round(cer / len(self), 6)

    def save_binary(
        self,
        file: str | Path,
        protocol: Literal[
            "protobuf", "pickle", "json", "json-array", "protobuf-array", "pickle-array"
        ] = "protobuf",
        compress: Literal["lz4", "bz2", "lzma", "zlib", "gzip"] | None = None,
        show_progress: bool = False,
        clear_x_waveform: bool = True,
    ) -> None:
        """Save the audio list to a binary file.

        Args:
            file (str | Path): Path to the binary file.
            protocol (Literal[ &quot;protobuf&quot;, &quot;pickle&quot;, &quot;json&quot;, &quot;json, optional): the protocol to use. Defaults to &quot;protobuf&quot;.
            compress (Literal[&quot;lz4&quot;, &quot;bz2&quot;, &quot;lzma&quot;, &quot;zlib&quot;, &quot;gzip&quot;] | None, optional): the compression method to use. Defaults to None.
            show_progress (bool, optional): whether to show progress bar. Defaults to False.
            filter_bad (bool, optional): whether to filter bad audio files. Defaults to False.
            clear_x_waveform: whether clear the waveform of the input audio. Defaults to True.
        """
        for example in self:
            example: AudioEvalExample
            x_texts = []
            for channel in example.x.channels:
                channel: AudioChannel
                x_texts.append(channel.text)
            if clear_x_waveform:
                example.x.clear_waveform()
            example.x.clear_text()
            for i, channel in enumerate(example.x.channels):
                channel.raw_text = x_texts[i]
        return super().save_binary(file, protocol, compress, show_progress)

    @classmethod
    def load_binary(
        cls,
        file: str | bytes | Path,
        protocol: Literal[
            "protobuf", "pickle", "json", "json-array", "protobuf-array", "pickle-array"
        ] = "protobuf",
        compress: Literal["lz4", "bz2", "lzma", "zlib", "gzip"] | None = None,
        show_progress: bool = False,
        streaming: bool = False,
    ) -> AudioEvalResult | Generator[AudioEvalExample, None, None]:
        """Load the audio list from a binary file.

        Args:
            file (str | Path): Path to the binary file.
            show_progress (bool, optional): whether to show progress bar. Defaults to False.

        Returns:
            AudioEvalResult: Audio eval result.
        """
        docs = DocList[AudioEvalExample].load_binary(
            file=file,
            protocol=protocol,
            compress=compress,
            show_progress=show_progress,
            streaming=streaming,
        )
        if isinstance(docs, Generator):
            return docs
        return AudioEvalResult(docs=docs)

    def filter_by_cer(
        self,
        min_cer: float = 0.0,
        ignore_space: bool = True,
        ignore_punc: bool = True,
        normalize: bool = False,
    ) -> AudioEvalResult:
        """Filter examples whose character error rate is greater than min_cer.

        Args:
            min_cer: CER threshold, only examples with CER > min_cer are kept.
            ignore_space: Whether to ignore space when computing CER.
            ignore_punc: Whether to ignore punctuation when computing CER.
            normalize: Whether to normalize the text when computing CER.
        """
        filtered = []
        for example in self:
            example: AudioEvalExample
            cer = example.char_error_rate(
                ignore_space=ignore_space,
                ignore_punc=ignore_punc,
                normalize=normalize,
            )
            if cer > min_cer:
                filtered.append(example)
        return AudioEvalResult(docs=filtered)

    def get_badcase(self) -> AudioEvalResult:
        query = {"is_badcase": {"$eq": True}}
        return AudioEvalResult(docs=filter_docs(docs=self, query=query))



from functools import wraps
import time
from typing import Callable

from loguru import logger


def timer(
    tag: str,
    round_num: int = 2,
    is_class: bool = True,
    log_time: bool = True,
    accumulate_time: bool = False,
):
    def outter(func: Callable):
        if is_class:

            @wraps(func)
            def wrapper(self, *args, **kwargs):
                start = time.perf_counter()
                res = func(self, *args, **kwargs)
                end = time.perf_counter()
                spent = round(end - start, round_num)
                if log_time:
                    logger.info(f"{tag} run in {spent} seconds")
                if hasattr(self, "timer_data"):
                    if accumulate_time:
                        if self.timer_data is None:
                            self.timer_data = {}
                        if tag in self.timer_data:
                            self.timer_data[tag] += spent
                        else:
                            self.timer_data[tag] = spent
                    else:
                        self.timer_data[tag] = spent
                return res

            return wrapper

        else:

            @wraps(func)
            def wrapper(*args, **kwargs):
                start = time.perf_counter()
                res = func(*args, **kwargs)
                end = time.perf_counter()
                spent = round(end - start, round_num)
                if log_time:
                    logger.info(f"{tag} run in {spent} seconds")
                return res

            return wrapper

    return outter



import asyncio
from io import BytesIO
from pathlib import Path
from typing import List

import aiofiles
from aiohttp import ClientSession
import librosa
from loguru import logger
import numpy as np
from pydantic import Field
from pydub import AudioSegment

from fasr.config import Config, registry
from fasr.data.audio import Audio, AudioChannel, AudioChannelList, AudioList
from fasr.data.waveform import Waveform

from .base import BaseComponent

LIBROSA_AVAILABLE_FORMATS = ["wav", "mp3"]


@registry.components.register("loader")
class AudioLoader(BaseComponent):
    """异步音频下载器，负责所有音频的并行下载和下载条件"""

    name: str = "loader"
    input_tags: List[str] = ["Audio.url"]
    output_tags: List[str] = [
        "Audio.waveform",
        "Audio.duration",
        "Audio.channels",
    ]

    max_duration_seconds: float | None = Field(
        None, alias="max_duration", description="音频最大时长，超过该时长则截断"
    )
    min_duration_seconds: float | None = Field(
        None, alias="min_duration", description="音频最小时长，小于该时长则不下载"
    )
    reload: bool = Field(False, description="是否重新下载")
    only_num_channels: int | None = Field(
        None, description="只下载指定通道数的音频，None表示不限制"
    )
    mono: bool | None = Field(None, description="是否合并多通道音频为单通道")

    def predict(self, audios: AudioList) -> AudioList:
        _ = asyncio.run(self.aload_audios(audios=audios))
        return audios

    async def aload_audio(self, audio: Audio, session: ClientSession) -> Audio:
        if not self.reload and audio.is_loaded:
            return audio
        mono = self.mono if audio.mono is None else audio.mono
        audio_format = get_audio_format(audio.url)
        try:
            if Path(audio.url).exists():
                try:
                    async with aiofiles.open(audio.url, "rb") as f:
                        bytes = await f.read()
                        bytes = BytesIO(bytes)
                        if audio_format not in LIBROSA_AVAILABLE_FORMATS:
                            bytes = convert_bytes(
                                bytes, target_format="wav", ori_format=audio_format
                            )
                        try:
                            data, sample_rate = librosa.load(
                                bytes, sr=audio.sample_rate, mono=mono
                            )
                        except Exception:
                            logger.warning(
                                f"librosa can't load audio, using pydub to convert audio format, url: {audio.url}"
                            )
                            bytes = convert_bytes(bytes, "wav", ori_format=None)
                            data, sample_rate = librosa.load(
                                bytes, sr=audio.sample_rate, mono=mono
                            )
                        duration = librosa.get_duration(y=data, sr=sample_rate)
                        audio.duration = round(duration, 4)
                        audio = assign_audio_data(audio, data, sample_rate)
                        audio = check_audio_duration(
                            audio, self.max_duration_seconds, self.min_duration_seconds
                        )
                        audio = check_audio_channels(audio, self.only_num_channels)
                except Exception as e:
                    audio.is_bad = True
                    audio.bad_reason = str(e)
                    audio.bad_component = self.name

                return audio

            async with session.get(audio.url) as response:
                if response.status == 200:
                    bytes = await response.read()
                    bytes = BytesIO(bytes)
                    if audio_format not in LIBROSA_AVAILABLE_FORMATS:
                        bytes = convert_bytes(
                            bytes, target_format="wav", ori_format=audio_format
                        )
                    try:
                        data, sample_rate = librosa.load(
                            bytes, sr=audio.sample_rate, mono=mono
                        )
                    except Exception:
                        logger.warning(
                            f"librosa can't load audio, using pydub to convert audio format, url: {audio.url}"
                        )
                        bytes = await response.read()
                        bytes = BytesIO(bytes)
                        bytes = convert_bytes(bytes, "wav", ori_format=None)
                        data, sample_rate = librosa.load(
                            bytes, sr=audio.sample_rate, mono=mono
                        )
                    duration = librosa.get_duration(y=data, sr=sample_rate)
                    audio.duration = round(duration, 4)
                    audio = assign_audio_data(audio, data, sample_rate)
                    audio = check_audio_duration(
                        audio, self.max_duration_seconds, self.min_duration_seconds
                    )
                    audio = check_audio_channels(audio, self.only_num_channels)
                else:
                    audio.is_bad = True
                    audio.bad_reason = f"下载音频失败，状态码{response.status}"
                    audio.bad_component = self.name

                return audio
        except Exception as e:
            audio.is_bad = True
            audio.bad_reason = str(e)
            audio.bad_component = self.name
            return audio

    async def aload_audios(self, audios: AudioList) -> AudioList:
        async with ClientSession() as session:
            tasks = [self.aload_audio(audio, session) for audio in audios]
            results = await asyncio.gather(*tasks)
            return results

    def get_config(self) -> Config:
        data = {
            "component": {
                "@components": "loader.v2",
                "max_duration": self.max_duration_seconds,
                "min_duration": self.min_duration_seconds,
                "reload": self.reload,
                "mono": self.mono,
                "only_num_channels": self.only_num_channels,
            }
        }
        return Config(data=data)

    def save(self, save_dir: str | Path) -> None:
        save_dir = Path(save_dir)
        if not save_dir.exists():
            save_dir.mkdir(parents=True)
        config_path = save_dir / "config.cfg"
        self.get_config().to_disk(config_path)

    def load(self, save_dir: str | Path) -> "AudioLoader":
        config_path = Path(save_dir, "config.cfg")
        config = Config().from_disk(config_path)
        loader = registry.resolve(config)["component"]
        return loader

    def setup(self, checkpoint_dir, device="cpu", compile=False, **kwargs):
        pass


def convert_bytes(
    bytes: BytesIO, target_format: str, ori_format: str | None = None
) -> BytesIO:
    """using pydub to convert bytes to another format bytes.

    Args:
        bytes (BytesIO): The bytes to be converted.
        target_format (str): The target format.
        ori_format (str | None, optional): The original format. Defaults to None.

    Returns:
        BytesIO: The converted bytes.
    """
    segment: AudioSegment = AudioSegment.from_file(bytes, format=ori_format)
    bytes = BytesIO()
    segment.export(bytes, format=target_format)
    return bytes


def get_audio_format(url: str) -> str | None:
    """Get audio format from url.

    Args:
        url (str): The url of audio.

    Returns:
        str: The format of audio.
    """
    formats = [".mp3", ".wav", ".amr", ".m4a", ".flac", ".aac", ".ogg"]
    for format in formats:
        if format in url:
            return format.replace(".", "")
    return None


def assign_audio_data(audio: Audio, data: np.ndarray, sample_rate: int) -> Audio:
    """Assign audio data to audio.

    Args:
        audio (Audio): The audio to be assigned.
        data (np.ndarray): The audio data.
        sample_rate (int): The sample rate of audio.

    Returns:
        Audio: The audio with data assigned.
    """
    if len(data.shape) == 1:
        audio.mono = True
        audio.channels = AudioChannelList(
            [
                AudioChannel(
                    id=audio.id,
                    waveform=Waveform(data=data, sample_rate=sample_rate),
                )
            ]
        )
    else:
        audio.mono = False
        audio.channels = AudioChannelList(
            [
                AudioChannel(
                    id=audio.id,
                    waveform=Waveform(data=channel_data, sample_rate=sample_rate),
                )
                for channel_data in data
            ]
        )
    return audio


def check_audio_duration(
    audio: Audio,
    max_duration_seconds: float | None = None,
    min_duration_seconds: float | None = None,
) -> Audio:
    """Check audio duration.

    Args:
        audio (Audio): The audio to be checked.
        max_duration (float | None): The max duration of audio.
        min_duration (float | None): The min duration of audio.

    Returns:
        Audio: The audio with duration checked.
    """
    if max_duration_seconds and audio.duration > max_duration_seconds:
        audio.is_bad = True
        audio.bad_reason = f"音频时长超过最大时长限制{max_duration_seconds}s, 当前时长{audio.duration}s"
    if min_duration_seconds and audio.duration < min_duration_seconds:
        audio.is_bad = True
        audio.bad_reason = f"音频时长小于最小时长限制{min_duration_seconds}s, 当前时长{audio.duration}s"

    if audio.is_bad:
        audio.bad_component = "loader"
    return audio


def check_audio_channels(audio: Audio, only_num_channels: int | None = None) -> Audio:
    """Check audio channels.

    Args:
        audio (Audio): The audio to be checked.
        only_num_channels (int | None): The number of channels.

    Returns:
        Audio: The audio with channels checked.
    """
    if only_num_channels and len(audio.channels) != only_num_channels:
        audio.is_bad = True
        audio.bad_reason = f"音频通道数不符合要求, 期望{only_num_channels}通道，实际{len(audio.channels)}通道"
    if audio.is_bad:
        audio.bad_component = "loader"
    return audio



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Kai Huang)

from collections import deque
import enum
import logging

import numpy as np

from .constants import FRAME_LENGTH_MS, FRAME_LENGTH_S, FRAME_SHIFT_MS, FRAME_SHIFT_S

logger = logging.getLogger(__name__)


@enum.unique
class VadState(enum.Enum):
    SILENCE = 0
    POSSIBLE_SPEECH = 1
    SPEECH = 2
    POSSIBLE_SILENCE = 3


class VadPostprocessor:
    def __init__(
        self,
        smooth_window_size,
        prob_threshold,
        min_speech_frame,
        max_speech_frame,
        min_silence_frame,
        merge_silence_frame,
        extend_speech_frame,
    ):
        self.smooth_window_size = max(1, smooth_window_size)
        self.prob_threshold = prob_threshold
        self.min_speech_frame = min_speech_frame
        self.max_speech_frame = max_speech_frame
        self.min_silence_frame = min_silence_frame
        self.merge_silence_frame = merge_silence_frame
        self.extend_speech_frame = extend_speech_frame

    def process(self, raw_probs):
        if not raw_probs:
            return []

        smoothed_probs = self._smooth_prob(raw_probs)

        binary_preds = self._apply_threshold(smoothed_probs)

        # decision: 0 means silence, 1 means speech
        decisions = self._smooth_preds_with_state_machine(binary_preds)

        fixed_decisions = self._fix_smooth_window_start(decisions)
        smoothed_decisions = self._merge_short_silence_segments(fixed_decisions)
        extend_decisions = self._extend_speech_segments(smoothed_decisions)
        final_decisions = self._split_long_speech_segments(extend_decisions, raw_probs)
        # don't call _merge_short_silence_segments after _split_long_speech_segments

        return final_decisions

    def decision_to_segment(self, decisions, wav_dur=None):
        segments = []
        speech_start = None
        for t, decision in enumerate(decisions):
            if decision == 1 and speech_start is None:
                speech_start = t
            elif decision == 0 and speech_start is not None:
                if (t - speech_start) < self.min_speech_frame:
                    logger.warning(
                        "Unexpected short speech segment, check vad_postprocessor.py"
                    )
                segments.append((speech_start * FRAME_SHIFT_S, t * FRAME_SHIFT_S))
                speech_start = None
        if speech_start is not None:
            t = len(decisions) - 1
            if (t - speech_start) < self.min_speech_frame:
                logger.warning(
                    "Unexpected short speech segment, check vad_postprocessor.py"
                )
            end_time = len(decisions) * FRAME_SHIFT_S + FRAME_LENGTH_S
            if wav_dur is not None:
                end_time = min(end_time, wav_dur)
            segments.append((speech_start * FRAME_SHIFT_S, end_time))
        segments = [(round(s, 3), round(e, 3)) for s, e in segments]
        return segments

    def _smooth_prob_simple(self, probs):
        if self.smooth_window_size <= 1:
            return probs
        smoothed_probs = probs.copy()
        window = deque()
        window_sum = 0.0
        for i, p in enumerate(probs):
            window.append(p)
            window_sum += p
            if len(window) > self.smooth_window_size:
                left = window.popleft()
                window_sum -= left
            window_avg = window_sum / len(window)
            smoothed_probs[i] = window_avg
        return smoothed_probs

    def _smooth_prob(self, probs):
        if self.smooth_window_size <= 1:
            return np.asarray(probs)
        probs_np = np.array(probs)
        kernel = np.ones(self.smooth_window_size) / self.smooth_window_size
        # mode='same' 保持长度，'valid' 会变短
        smoothed = np.convolve(probs_np, kernel, mode="full")[: len(probs)]
        # 处理边界：前几帧用累积平均
        for i in range(min(self.smooth_window_size - 1, len(probs))):
            smoothed[i] = np.mean(probs_np[: i + 1])
        return smoothed  # .tolist()

    def _apply_threshold_simple(self, probs):
        return [int(p >= self.prob_threshold) for p in probs]

    def _apply_threshold(self, probs):
        probs_np = np.asarray(probs)
        return (probs_np >= self.prob_threshold).astype(int).tolist()

    def _smooth_preds_with_state_machine(self, binary_preds):
        """
        state transition is constrained by min_speech_frame & min_silence_frame
        """
        if self.min_speech_frame <= 0 and self.min_silence_frame <= 0:
            return binary_preds
        decisions = [0] * len(binary_preds)
        state = VadState.SILENCE
        speech_start = -1
        silence_start = -1
        for t, is_speech in enumerate(binary_preds):
            # State transition
            if state == VadState.SILENCE:
                if is_speech:
                    state = VadState.POSSIBLE_SPEECH
                    speech_start = t

            elif state == VadState.POSSIBLE_SPEECH:
                if is_speech:
                    assert speech_start != -1
                    if t - speech_start >= self.min_speech_frame:
                        state = VadState.SPEECH
                        decisions[speech_start:t] = [1] * (t - speech_start)
                else:
                    state = VadState.SILENCE
                    speech_start = -1

            elif state == VadState.SPEECH:
                if not is_speech:
                    state = VadState.POSSIBLE_SILENCE
                    silence_start = t

            elif state == VadState.POSSIBLE_SILENCE:
                if not is_speech:
                    assert silence_start != -1
                    if t - silence_start >= self.min_silence_frame:
                        state = VadState.SILENCE
                        speech_start = -1
                else:
                    state = VadState.SPEECH
                    silence_start = -1

            # current frame's decision
            if state == VadState.SPEECH or state == VadState.POSSIBLE_SILENCE:
                decision = 1
            elif state == VadState.SILENCE or state == VadState.POSSIBLE_SPEECH:
                decision = 0
            else:
                raise ValueError("Impossible VAD state")

            decisions[t] = decision
        return decisions

    def _fix_smooth_window_start(self, decisions):
        new_decisions = decisions.copy()
        for t, decision in enumerate(decisions):
            if t > 0 and decisions[t - 1] == 0 and decision == 1:
                start = max(0, t - self.smooth_window_size)
                new_decisions[start:t] = [1] * (t - start)
        return new_decisions

    def _merge_short_silence_segments(self, decisions):
        if self.merge_silence_frame <= 0:
            return decisions
        new_decisions = decisions.copy()
        silence_start = None
        for t, decision in enumerate(decisions):
            if (
                t > 0
                and decisions[t - 1] == 1
                and decision == 0
                and silence_start is None
            ):
                silence_start = t
            elif (
                t > 0
                and decisions[t - 1] == 0
                and decision == 1
                and silence_start is not None
            ):
                silence_frame = t - silence_start
                if silence_frame < self.merge_silence_frame:
                    new_decisions[silence_start:t] = [1] * silence_frame
                silence_start = None
        return new_decisions

    def _extend_speech_segments_simple(self, decisions):
        """
        extend N frames before & after speech segments
        """
        if self.extend_speech_frame <= 0:
            return decisions
        new_decisions = decisions.copy()
        for t, decision in enumerate(decisions):
            if decision == 1:
                start = max(0, t - self.extend_speech_frame)
                end = min(len(decisions), t + self.extend_speech_frame + 1)
                new_decisions[start:end] = [1] * (end - start)
        return new_decisions

    def _extend_speech_segments(self, decisions):
        """
        extend N frames before & after speech segments
        """
        if self.extend_speech_frame <= 0:
            return decisions
        decisions_np = np.array(decisions)
        kernel = np.ones(2 * self.extend_speech_frame + 1)
        extended = np.convolve(decisions_np, kernel, mode="same")
        return (extended > 0).astype(int).tolist()

    def _split_long_speech_segments(self, decisions, probs):
        new_decisions = decisions.copy()
        segments = self.decision_to_segment(decisions)
        for start_s, end_s in segments:
            start_frame = int(start_s / FRAME_SHIFT_S)
            end_frame = int(end_s / FRAME_SHIFT_S)
            dur_frames = end_frame - start_frame
            if dur_frames > self.max_speech_frame:
                segment_probs = probs[start_frame:end_frame]
                split_points = self._find_split_points(segment_probs)
                for split_point in split_points:
                    split_frame = start_frame + split_point
                    new_decisions[split_frame] = 0
        return new_decisions

    def _find_split_points(self, probs):
        split_points = []
        length = len(probs)
        start = 0
        while start < length:
            if (length - start) <= self.max_speech_frame:
                break
            window_start = int(start + self.max_speech_frame / 2)
            window_end = int(start + self.max_speech_frame)
            window_probs = probs[window_start:window_end]

            min_index = window_start + np.argmin(window_probs)
            split_points.append(min_index)

            start = min_index + 1
        return split_points



import numpy as np
import pytest

from fasr.data.audio import (
    Audio,
    AudioChannel,
    AudioChannelList,
    AudioChunk,
    AudioChunkList,
    AudioSpan,
    AudioSpanList,
    AudioToken,
    AudioTokenList,
)
from fasr.data.waveform import Waveform


@pytest.fixture
def sample_rate():
    return 16000


@pytest.fixture
def sine_wave(sample_rate):
    """1 秒 440Hz 正弦波"""
    t = np.linspace(0, 1, sample_rate, dtype=np.float32)
    return np.sin(2 * np.pi * 440 * t)


@pytest.fixture
def short_wave(sample_rate):
    """0.5 秒正弦波"""
    t = np.linspace(0, 0.5, sample_rate // 2, dtype=np.float32)
    return np.sin(2 * np.pi * 440 * t)


@pytest.fixture
def waveform(sine_wave, sample_rate):
    return Waveform(data=sine_wave, sample_rate=sample_rate)


@pytest.fixture
def short_waveform(short_wave, sample_rate):
    return Waveform(data=short_wave, sample_rate=sample_rate)


@pytest.fixture
def audio_token():
    return AudioToken(text="你", start_ms=0, end_ms=100, follow="")


@pytest.fixture
def audio_token_list():
    return AudioTokenList(
        [
            AudioToken(text="你", start_ms=0, end_ms=100, follow=""),
            AudioToken(text="好", start_ms=100, end_ms=200, follow=""),
            AudioToken(text="世", start_ms=200, end_ms=300, follow=""),
            AudioToken(text="界", start_ms=300, end_ms=400, follow=" "),
        ]
    )


@pytest.fixture
def audio_span(audio_token_list, waveform):
    return AudioSpan(
        start_ms=0,
        end_ms=1000,
        waveform=waveform,
        tokens=audio_token_list,
    )


@pytest.fixture
def audio_channel(waveform, audio_token_list):
    segments = AudioSpanList(
        [
            AudioSpan(
                start_ms=0,
                end_ms=500,
                waveform=waveform,
                tokens=audio_token_list,
            )
        ]
    )
    return AudioChannel(waveform=waveform, segments=segments)


@pytest.fixture
def audio_with_channel(waveform, sample_rate):
    channel = AudioChannel(waveform=waveform)
    audio = Audio(sample_rate=sample_rate)
    audio.channels = AudioChannelList([channel])
    audio.duration = waveform.duration
    return audio


@pytest.fixture
def audio_chunk(waveform):
    return AudioChunk(stream_id="test-stream-id", waveform=waveform)



#!/usr/bin/env python3

# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

import argparse
import json
import logging
import os

from fireredasr2.asr import FireRedAsr2, FireRedAsr2Config
from fireredasr2.utils.io import get_wav_info, write_textgrid

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s (%(module)s:%(lineno)d) %(levelname)s: %(message)s",
)
logger = logging.getLogger("fireredasr2.bin.speech2text")


parser = argparse.ArgumentParser()
parser.add_argument("--asr_type", type=str, required=True, choices=["aed", "llm"])
parser.add_argument("--model_dir", type=str, required=True)

# Input / Output
parser.add_argument("--wav_path", type=str)
parser.add_argument("--wav_paths", type=str, nargs="*")
parser.add_argument("--wav_dir", type=str)
parser.add_argument("--wav_scp", type=str)
parser.add_argument("--sort_wav_by_dur", type=int, default=0)
parser.add_argument("--output", type=str)

# Decode Options
parser.add_argument("--use_gpu", type=int, default=1)
parser.add_argument("--use_half", type=int, default=0)
parser.add_argument("--batch_size", type=int, default=1)
parser.add_argument("--beam_size", type=int, default=1)
parser.add_argument("--decode_max_len", type=int, default=0)
# FireRedASR-AED
parser.add_argument("--nbest", type=int, default=1)
parser.add_argument("--softmax_smoothing", type=float, default=1.0)
parser.add_argument("--aed_length_penalty", type=float, default=0.0)
parser.add_argument("--eos_penalty", type=float, default=1.0)
parser.add_argument("--return_timestamp", type=int, default=0)
parser.add_argument("--write_textgrid", type=int, default=0)
# AED External LM
parser.add_argument("--elm_dir", type=str, default="")
parser.add_argument("--elm_weight", type=float, default=0.0)
# FireRedASR-LLM
parser.add_argument("--decode_min_len", type=int, default=0)
parser.add_argument("--repetition_penalty", type=float, default=1.0)
parser.add_argument("--llm_length_penalty", type=float, default=0.0)
parser.add_argument("--temperature", type=float, default=1.0)


def main(args):
    wavs = get_wav_info(args)
    fout = open(args.output, "w") if args.output else None
    foutl = open(args.output + ".jsonl", "w") if args.output else None

    asr_config = FireRedAsr2Config(
        args.use_gpu,
        args.use_half,
        args.beam_size,
        args.nbest,
        args.decode_max_len,
        args.softmax_smoothing,
        args.aed_length_penalty,
        args.eos_penalty,
        args.return_timestamp,
        args.decode_min_len,
        args.repetition_penalty,
        args.llm_length_penalty,
        args.temperature,
        args.elm_dir,
        args.elm_weight,
    )
    model = FireRedAsr2.from_pretrained(args.asr_type, args.model_dir, asr_config)

    batch_uttid = []
    batch_wav_path = []
    for i, wav in enumerate(wavs):
        uttid, wav_path = wav
        batch_uttid.append(uttid)
        batch_wav_path.append(wav_path)
        if len(batch_wav_path) < args.batch_size and i != len(wavs) - 1:
            continue

        results = model.transcribe(batch_uttid, batch_wav_path)

        for result in results:
            logger.info(result)
            if fout is not None:
                foutl.write(f"{json.dumps(result, ensure_ascii=False)}\n")
                fout.write(f"{result['uttid']}\t{result['text']}\n")
            if args.write_textgrid and "timestamp" in result:
                write_textgrid(result["wav"], result["dur_s"], result["timestamp"])

        if fout:
            fout.flush()
        if foutl:
            foutl.flush()
        batch_uttid = []
        batch_wav_path = []
    if fout:
        fout.close()
    if foutl:
        foutl.close()


if __name__ == "__main__":
    args = parser.parse_args()
    logger.info(args)
    main(args)



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Kai Huang)

SAMPLE_RATE = 16000
FRAME_LENGTH_MS = 25
FRAME_SHIFT_MS = 10
FRAME_LENGTH_S = 0.025
FRAME_SHIFT_S = 0.010
FRAME_LENGTH_SAMPLE = int(SAMPLE_RATE * FRAME_LENGTH_MS / 1000)
FRAME_SHIFT_SAMPLE = int(SAMPLE_RATE * FRAME_SHIFT_MS / 1000)
FRAME_PER_SECONDS = int(1000 / FRAME_SHIFT_MS)



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

import traceback

import torch
import torchaudio

from .module.conformer_encoder import ConformerEncoder
from .module.ctc import CTC
from .module.transformer_decoder import TransformerDecoder


class FireRedAsrAed(torch.nn.Module):
    @classmethod
    def from_args(cls, args):
        return cls(args)

    def __init__(self, args):
        super().__init__()
        self.sos_id = args.sos_id
        self.eos_id = args.eos_id

        self.encoder = ConformerEncoder(
            args.idim,
            args.n_layers_enc,
            args.n_head,
            args.d_model,
            args.residual_dropout,
            args.dropout_rate,
            args.kernel_size,
            args.pe_maxlen,
        )

        self.decoder = TransformerDecoder(
            args.sos_id,
            args.eos_id,
            args.pad_id,
            args.odim,
            args.n_layers_dec,
            args.n_head,
            args.d_model,
            args.residual_dropout,
            args.pe_maxlen,
        )

        self.ctc = CTC(args.odim, args.d_model)

    def transcribe(
        self,
        padded_input,
        input_lengths,
        beam_size=1,
        nbest=1,
        decode_max_len=0,
        softmax_smoothing=1.0,
        length_penalty=0.0,
        eos_penalty=1.0,
        return_timestamp=False,
        elm=None,
        elm_weight=0.0,
    ):
        enc_outputs, enc_lengths, enc_mask = self.encoder(padded_input, input_lengths)
        nbest_hyps = self.decoder.batch_beam_search(
            enc_outputs,
            enc_mask,
            beam_size,
            nbest,
            decode_max_len,
            softmax_smoothing,
            length_penalty,
            eos_penalty,
            elm,
            elm_weight,
        )
        if return_timestamp:
            nbest_hyps = self.get_token_timestamp_torchaudio(
                enc_outputs, enc_lengths, nbest_hyps
            )
        return nbest_hyps

    def get_token_timestamp_torchaudio(self, enc_outputs, enc_lengths, nbest_hyps):
        ctc_logits = self.ctc(enc_outputs)
        enc_lengths = enc_lengths
        for n in range(enc_outputs.size(0)):
            try:
                n_ctc_logits = ctc_logits[n, : enc_lengths[n]]
                y = nbest_hyps[n][0]["yseq"]
                y = y[y != 0]  # 0 is blank
                if y.numel() == 0 or n_ctc_logits.size()[0] == 0:
                    logger.debug("skip null output")
                    nbest_hyps[n][0]["timestamp"] = None
                    continue
                elif y.numel() > n_ctc_logits.size()[0]:
                    nbest_hyps[n][0]["timestamp"] = None
                    continue

                alignment, _ = torchaudio.functional.forced_align(
                    n_ctc_logits.unsqueeze(0), y.unsqueeze(0), blank=0
                )
                alignment = alignment[0].cpu().tolist()
                start_times, end_times = self.ctc.ctc_alignment_to_timestamp(
                    alignment, self.encoder.input_preprocessor.subsampling, blank_id=0
                )
                nbest_hyps[n][0]["timestamp"] = (start_times, end_times)
            except:
                traceback.print_exc()
                nbest_hyps[n][0]["timestamp"] = None
        return nbest_hyps



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

import numpy as np
import torch
import torch.nn.functional as F


class CTC(torch.nn.Module):
    def __init__(self, odim, encoder_output_size):
        super().__init__()
        self.ctc_lo = torch.nn.Linear(encoder_output_size, odim)

    def forward(self, encoder_output_pad):
        """encoder_output_pad: (N, T, H)"""
        return F.log_softmax(self.ctc_lo(encoder_output_pad), dim=2)

    @classmethod
    def ctc_align(cls, ctc_probs, y, blank_id=0):
        """ctc forced alignment.

        Args:
            torch.Tensor ctc_probs: hidden state sequence, 2d tensor (T, D)
            torch.Tensor y: id sequence tensor 1d tensor (L)
            int blank_id: blank symbol index
        Returns:
            torch.Tensor: alignment result
        """
        y_insert_blank = insert_blank(y, blank_id)

        log_alpha = torch.zeros((ctc_probs.size(0), len(y_insert_blank)))
        log_alpha = log_alpha - float("inf")  # log of zero
        state_path = (
            torch.zeros((ctc_probs.size(0), len(y_insert_blank)), dtype=torch.int16) - 1
        )  # state path

        # init start state
        log_alpha[0, 0] = ctc_probs[0][y_insert_blank[0]]
        log_alpha[0, 1] = ctc_probs[0][y_insert_blank[1]]

        for t in range(1, ctc_probs.size(0)):
            for s in range(len(y_insert_blank)):
                if (
                    y_insert_blank[s] == blank_id
                    or s < 2
                    or y_insert_blank[s] == y_insert_blank[s - 2]
                ):
                    candidates = torch.tensor(
                        [log_alpha[t - 1, s], log_alpha[t - 1, s - 1]]
                    )
                    prev_state = [s, s - 1]
                else:
                    candidates = torch.tensor(
                        [
                            log_alpha[t - 1, s],
                            log_alpha[t - 1, s - 1],
                            log_alpha[t - 1, s - 2],
                        ]
                    )
                    prev_state = [s, s - 1, s - 2]
                log_alpha[t, s] = (
                    torch.max(candidates) + ctc_probs[t][y_insert_blank[s]]
                )
                state_path[t, s] = prev_state[torch.argmax(candidates)]

        state_seq = -1 * torch.ones((ctc_probs.size(0), 1), dtype=torch.int16)

        candidates = torch.tensor(
            [
                log_alpha[-1, len(y_insert_blank) - 1],
                log_alpha[-1, len(y_insert_blank) - 2],
            ]
        )
        prev_state = [len(y_insert_blank) - 1, len(y_insert_blank) - 2]
        state_seq[-1] = prev_state[torch.argmax(candidates)]
        for t in range(ctc_probs.size(0) - 2, -1, -1):
            state_seq[t] = state_path[t + 1, state_seq[t + 1, 0]]

        output_alignment = []
        for t in range(0, ctc_probs.size(0)):
            output_alignment.append(y_insert_blank[state_seq[t, 0]])

        return output_alignment

    @classmethod
    def ctc_alignment_to_timestamp(cls, ys_with_blank, subsampling, blank_id=0):
        start_times: List[float] = []
        end_times: List[float] = []
        frame_shift = 10  # ms, hard code
        T = len(ys_with_blank)
        t = 0
        ctc_durs = []
        while t < T:
            token = ys_with_blank[t]
            t += 1
            if token != blank_id:
                start_t = t
                timestamp = frame_shift * subsampling * t / 1000.0  # s
                start_times.append(timestamp)
                if len(start_times) == len(end_times) + 2:
                    end_times.append(start_times[-1])
                # skip repeat token
                while t < T and token == ys_with_blank[t]:
                    t += 1
                assert t - start_t >= 0
                ctc_durs.append((t - start_t + 1) * frame_shift * subsampling / 1000.0)
        end_times.append((frame_shift * subsampling * T + 25) / 1000.0)
        if len(start_times) == 0:
            start_times.append(0.0)

        # Refine end_times
        assert len(ctc_durs) == len(end_times) and len(start_times) == len(end_times)
        avg_dur = sum(e - s for s, e in zip(start_times, end_times)) / len(end_times)
        new_end_times = []
        for s, e, ctc_dur in zip(start_times, end_times, ctc_durs):
            if e - s > 2 * avg_dur:
                e = s + max(1.5 * avg_dur, ctc_dur)
            new_end_times.append(round(e, 3))
        end_times = new_end_times
        return start_times, end_times


def insert_blank(label, blank_id=0):
    """Insert blank token between every two label token."""
    label = np.expand_dims(label, 1)
    blanks = np.zeros((label.shape[0], 1), dtype=np.int64) + blank_id
    label = np.concatenate([blanks, label], axis=1)
    label = label.reshape(-1)
    label = np.append(label, label[0])
    return label



from __future__ import annotations

from functools import lru_cache
from io import BytesIO
import os
from pathlib import Path
import re
from typing import (
    Any,
    Dict,
    FrozenSet,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Union,
)

from docarray import BaseDoc, DocList
from docarray.typing import ID, AnyTensor
from docarray.utils.filter import filter_docs
from joblib import Parallel, delayed
import librosa
from loguru import logger
import numpy as np
from pydantic import ConfigDict, Field, model_validator
from pydub import AudioSegment
import requests
from typing_extensions import Self

from .waveform import Waveform


class AudioToken(BaseDoc):
    """Audio token object that represents the smallest unit of an audio file, which must have the text that the speech recognizer recognized.

    Args:
        start_ms (int): Start ms of the token.
        end_ms (int): End ms of the token.
        text (str): Text of the token.
        waveform (NdArray): Waveform of the token.
        follow (str): Follow char. Defaults to " ".
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: ID | None = None
    text: str = Field(..., description="Text of the token.")
    start_ms: int | None = Field(None, description="Start time of the token.")
    end_ms: int | None = Field(None, description="End time of the token.")
    waveform: Waveform | None = Field(
        None, description="Waveform of the token.", exclude=True
    )
    follow: Optional[str] = Field(" ", description="Follow char.")

    @property
    def duration_ms(self):
        """Get the duration of the token in milliseconds."""
        return self.end_ms - self.start_ms

    def is_punctuation(self) -> bool:
        return re.match(r"[^\w\s]", self.text) is not None


class AudioTokenList(DocList[AudioToken]):
    @property
    def duration_ms(self):
        """Get the total duration of all the tokens in milliseconds."""
        return sum([token.duration_ms for token in self])

    @property
    def duration(self):
        """Get the total duration of all the tokens in seconds."""
        return self.duration_ms / 1000

    @property
    def text(self):
        """Get the text of all the tokens."""
        text = ""
        for t in self:
            text += t.text
            text += t.follow
        return text.strip()  # remove the last space


class AudioSpan(BaseDoc):
    """Audio span object that represents a segment of an audio file, which must have the start and end time.

    Args:
        start_ms (int): Start ms of the segment.
        end_ms (int): End ms of the segment.
        waveform (NdArray): Waveform of the segment.
        feats (NdArray): Features of the segment.
        tokens (AudioTokenList): Tokens of the segment.
        sample_rate (int): Sample rate of the segment.
        is_last (bool): Whether the segment is the last segment. Defaults to False.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="ignore")

    id: ID | None = None
    start_ms: float | int | None = Field(None, description="Start time of the segment.")
    end_ms: float | int | None = Field(None, description="End time of the segment.")
    waveform: Waveform | None = Field(
        None, description="Waveform of the segment.", exclude=True
    )

    feats: AnyTensor | None = Field(None, description="Features of the segment.")
    scores: AnyTensor | None = Field(None, description="Scores of the segment.")
    tokens: AudioTokenList | None = Field(None, description="Tokens of the segment.")

    language: str | None = None
    emotion: str | None = None
    type: str | None = None
    is_last: bool | None = None
    is_bad: bool | None = None
    bad_reason: str | None = None
    bad_component: str | None = None

    def display(self):
        """Display the segment."""
        from IPython.display import Audio as IPAudio

        return IPAudio(data=self.waveform.data, rate=self.sample_rate)

    @property
    def sample_rate(self) -> int | None:
        """Get the sample rate of the segment."""
        if self.waveform is None:
            return None
        return self.waveform.sample_rate

    @property
    def duration_ms(self):
        """Get the duration of the segment in milliseconds."""
        return self.end_ms - self.start_ms

    @property
    def duration(self):
        """Get the duration of the segment in seconds."""
        return self.duration_ms / 1000

    @property
    def text(self):
        """Get the text of all the tokens."""
        if hasattr(self, "_text"):
            return self._text
        text = ""
        if self.tokens is None:
            return text
        for t in self.tokens:
            text += t.text
            text += t.follow
        return text.strip()  # remove the last space

    @text.setter
    def text(self, text: str):
        self._text = text

    def __lt__(self, other: "AudioSpan") -> bool:
        """Compare the duration of the segment with another segment. like `self < other`.

        Args:
            other (AudioSpan): Another segment.

        Returns:
            bool: Whether the duration of the segment is less than the duration of the other segment.
        """
        return self.duration_ms < other.duration_ms

    def __gt__(self, other: "AudioSpan") -> bool:
        """Compare the duration of the segment with another segment. like `self > other`.

        Args:
            other (AudioSpan): Another segment.

        Returns:
            bool: Whether the duration of the segment is greater than the duration of the other segment.
        """
        return self.duration_ms > other.duration_ms

    def __getitem__(self, index: int) -> Optional[AudioToken]:
        """Get the token at the index.

        Args:
            index (int): Index of the token.

        Returns:
            AudioToken: Token at the index.
        """
        if self.tokens is None:
            return None
        return self.tokens[index]

    def __len__(self) -> int:
        """Get the number of tokens in the segment.

        Returns:
            int: Number of tokens in the segment.
        """
        if self.tokens is None:
            return 0
        return len(self.tokens)


class AudioSpanList(DocList[AudioSpan]):
    @property
    def duration_ms(self):
        """Get the total duration of all the spans in milliseconds."""
        return sum([span.duration_ms for span in self])

    @property
    def duration(self):
        """Get the total duration of all the spans in seconds."""
        return self.duration_ms / 1000

    @property
    def padded_duration_ms(self):
        """Get the padded duration of all the spans in milliseconds."""
        all_durations = [span.duration_ms for span in self]
        if len(all_durations) == 0:
            return 0
        duration = max(all_durations) * len(all_durations)
        return duration

    @property
    def max_duration_ms(self):
        """Get the maximum duration of all the spans in milliseconds."""
        all_durations = [span.duration_ms for span in self]
        if len(all_durations) == 0:
            return 0
        return max(all_durations)

    @property
    def text(self):
        """Get the text of all the spans."""
        _text = ""
        for span in self:
            _text += span.text
        return _text


class AudioChannel(BaseDoc):
    """Audio channel object that represents a channel of an audio file."""

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    id: ID | None = None
    raw_text: str | None = Field(None, description="Text of the channel.")
    waveform: Waveform | None = Field(
        None, description="Waveform of the audio file.", exclude=True
    )
    segments: Optional[AudioSpanList] = Field(
        None, description="Segments of the audio file provided by the detector."
    )
    sents: Optional[AudioSpanList] = Field(
        None, description="Sentences of the audio file provided by the sentencizer."
    )
    tokens: AudioTokenList | None = Field(
        default=None,
        description="Tokens of the audio file provided by the recognizer.",
    )
    stream: Iterable[AudioToken] | None = Field(
        None, description="Stream of tokens of the audio file."
    )
    is_last: Optional[bool] = Field(
        None, description="Whether the channel is the last."
    )
    language: str | None = Field(
        None, description="Language of the channel identified by LID."
    )
    is_bad: Optional[bool] = Field(None, description="Whether the channel is bad.")
    bad_reason: Optional[Union[str, Exception]] = Field(
        None, description="Reason why the audio file is bad."
    )

    @property
    def duration(self):
        """Get the duration of the audio file in seconds."""
        return self.waveform.duration

    @property
    def duration_s(self):
        """Get the duration of the audio file in seconds."""
        return self.duration

    @property
    def duration_ms(self):
        """Get the duration of the audio file in milliseconds."""
        return self.duration * 1000

    @property
    def start_ms(self):
        return 0

    @property
    def end_ms(self):
        return self.duration_ms

    @property
    def text(self):
        """Get the text of all the spans."""
        if self.raw_text is not None:
            return self.raw_text
        if self.sents:
            text = ""
            for sent in self.sents:
                text += sent.text
            return text
        elif self.segments:
            text = ""
            for seg in self.segments:
                text += seg.text
            return text
        elif self.tokens:
            text = ""
            for token in self.tokens:
                text += token.text + token.follow
            return text
        else:
            return ""

    @text.setter
    def text(self, text: str):
        self.raw_text = text

    def resample(self, sample_rate: int) -> "AudioChannel":
        """Resample the audio channel waveform to the target sample rate.

        Args:
            sample_rate (int): Target sample rate.
        """
        self.waveform = self.waveform.resample(sample_rate)
        return self

    def align_bad(self) -> bool:
        """Check if the audio channel is bad."""
        for span in self.segments:  # recognizer报错
            if span.is_bad:
                self.is_bad = True
                self.bad_reason = span.bad_reason
                return True
        for span in self.sents:  # sentencizer报错
            if span.is_bad:
                return True
        for span in self.steps:  # detector报错
            if span.is_bad:
                return True
        return False

    def is_recognized(self) -> bool:
        """Check if the audio channel is recognized."""
        if not self.segments:
            return False
        for span in self.segments:
            span: AudioSpan
            if span.tokens is None:
                return False
        return True

    def display(self, start_ms: float | None = None, end_ms: float | None = None):
        """Display the audio channel."""
        from IPython.display import Audio as IPAudio

        waveform = self.waveform.select_by_ms(start_ms, end_ms)

        return IPAudio(data=waveform.data, rate=waveform.sample_rate)

    def select_waveform(
        self, start_ms: float | None, end_ms: float | None = None
    ) -> np.ndarray:
        return self.waveform[start_ms:end_ms]

    def add_token(
        self,
        text: str,
        start_ms: int | None = None,
        end_ms: int | None = None,
        follow=" ",
    ) -> None:
        if self.tokens is None:
            self.tokens = AudioTokenList()
        if start_ms and end_ms:
            waveform = self.select_waveform(start_ms, end_ms)
        else:
            waveform = None
        self.tokens.append(
            AudioToken(
                start_ms=start_ms,
                end_ms=end_ms,
                text=text,
                waveform=waveform,
                follow=follow,
            )
        )

    def clear(self) -> None:
        self.waveform = None
        self.feats = None
        self.steps = None
        if self.segments:
            for span in self.segments:
                span.waveform = None
                span.feats = None
                span.scores = None
        if self.sents:
            for span in self.sents:
                span.waveform = None
                span.feats = None
                span.scores = None

    def set_waveform(self, data: np.ndarray, sample_rate: int) -> Self:
        self.waveform = Waveform(data=data, sample_rate=sample_rate)
        return self

    def set_text(self, text: str) -> Self:
        self.raw_text = text
        return self

    def __getitem__(self, index) -> Optional[AudioSpan]:
        if not self.segments:
            return None
        return self.segments[index]

    def __len__(self) -> int:
        if not self.segments:
            return 0
        return len(self.segments)


class AudioChannelList(DocList[AudioChannel]):
    @property
    def text(self):
        """Get the text of all the spans."""
        if not self.sents:
            return ""
        else:
            all_sents = [sent for channel in self for sent in channel.sents]
            sorted_sents = sorted(all_sents, key=lambda x: x.start_ms)
            text = ""
            for sent in sorted_sents:
                text += sent.text
            return text


class Audio(BaseDoc):
    """Audio object that represents an audio file.

    Args:
        url (HttpUrl): URL of the audio file.
        sample_rate (Optional[int], optional): Sample rate of the audio file. Defaults to None.
        waveform (Optional[NdArray], optional): Waveform of the audio file. Defaults to None.
        mono (Optional[bool], optional): Whether the audio file is mono. Defaults to None.
        feats (Optional[NdArray], optional): Features of the audio file. Defaults to None.
        duration (Optional[float], optional): Duration of the audio file. Defaults to None.
        segments (Union[Iterable[List], List], optional): Segments of the audio file. Defaults to None.
        pipeline (List[str], optional): List of processing steps. Defaults to [].
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True, validate_assignment=True, extra="ignore"
    )

    url: Union[str, Path] | None = Field(None, description="URL of the audio file.")
    sample_rate: int | None = Field(None, description="Sample rate of the audio file.")
    channels: AudioChannelList | None = Field(
        None, description="Channels of the audio file."
    )
    mono: bool | None = Field(None, description="Whether the audio file is mono.")
    duration: float | None = Field(None, description="Duration of the audio file.")
    pipeline: List[str] | None = Field(None, description="List of processing steps.")
    is_bad: bool | None = Field(None, description="Whether the audio file is bad.")
    hotwords: FrozenSet[str] | None = Field(
        None, description="Hotwords of the audio file."
    )
    bad_reason: str | Exception | None = Field(
        None, title="Reason why the audio file is bad."
    )
    bad_component: str | None = Field(
        None, title="Component that marked the audio bad."
    )
    spent_time: Dict[str, float] | None = Field(
        None, title="Time spent on processing the audio."
    )

    @model_validator(mode="after")
    def validate_audio(self):
        if isinstance(self.url, Path):
            self.url = str(
                self.url
            )  # convert Path to str, otherwise it will raise error when saving binary using protobuf protocol
        return self

    def load(self, mono: bool = False) -> "Audio":
        """Load the audio file from the URL."""
        try:
            if Path(self.url).exists():
                data, sample_rate = librosa.load(
                    self.url, sr=self.sample_rate, mono=mono
                )
            else:
                bytes_ = requests.get(self.url).content
                data, sample_rate = librosa.load(
                    BytesIO(bytes_), sr=self.sample_rate, mono=mono
                )
            self.duration = round(librosa.get_duration(y=data, sr=sample_rate), 4)
            self.sample_rate = sample_rate
        except Exception as e:
            raise ValueError(f"Failed to load audio from {self.url}. {e}")
        if len(data.shape) == 1:
            self.mono = True
            self.channels = AudioChannelList(
                [
                    AudioChannel(
                        id=self.id,
                        waveform=Waveform(
                            data=data, sample_rate=sample_rate, is_normalized=True
                        ),
                    )
                ]
            )

        else:
            self.mono = False
            self.channels = AudioChannelList(
                [
                    AudioChannel(
                        id=self.id,
                        waveform=Waveform(
                            data=channel_waveform_data,
                            sample_rate=sample_rate,
                            is_normalized=True,
                        ),
                    )
                    for channel_waveform_data in data
                ]
            )
        return self

    def load_example(self, example: Literal["vad", "asr"] = "asr") -> Self:
        vad_example_path = (
            Path(__file__).parent.parent / "asset" / "example" / "vad_example.wav"
        )
        asr_example_path = (
            Path(__file__).parent.parent / "asset" / "example" / "asr_example.wav"
        )
        if example == "vad":
            audio = Audio(url=vad_example_path).load()
            return audio
        elif example == "asr":
            audio = Audio(url=asr_example_path).load()
            return audio
        else:
            raise ValueError(f"Example {example} not found.")

    def load_bytes_io(self, bytes_io: BytesIO) -> Audio:
        """Load the audio file from the BytesIO object.

        Args:
            url (str): URL of the audio file or local path.
            bytes_io (BytesIO): BytesIO object.

        Returns:
            Audio: Audio object.
        """
        assert self.url is not None, "Audio URL is required."
        if self.url[-3:].lower() == "amr" or self.url[-3:].lower() == "m4a":
            audio: AudioSegment = AudioSegment.from_file(
                bytes_io, format=self.url[-3:].lower()
            )
            bytes_io = BytesIO()
            audio.export(bytes_io, format="wav")
        if not any(
            audio_format in self.url.lower()
            for audio_format in (".mp3", ".wav", ".amr", ".m4a")
        ):
            audio: AudioSegment = AudioSegment.from_file(bytes_io)
            bytes_io = BytesIO()
            audio.export(bytes_io, format="wav")
        try:
            arr, sr = librosa.load(bytes_io, sr=self.sample_rate, mono=self.mono)
        except Exception as e:
            logger.warning(f"librosa load failed: {e}, using pydub instead.")
            audio = AudioSegment.from_file(bytes_io)
            bytes_io = BytesIO()
            audio.export(bytes_io, format="wav")
            arr, sr = librosa.load(bytes_io, sr=self.sample_rate, mono=self.mono)
        if len(arr.shape) == 1:
            self.mono = True
            self.channels = AudioChannelList(
                [
                    AudioChannel(
                        id=self.id,
                        waveform=Waveform(data=arr, sample_rate=sr, is_normalized=True),
                    )
                ]
            )
        else:
            self.mono = False
            self.channels = AudioChannelList(
                [
                    AudioChannel(
                        id=self.id,
                        waveform=Waveform(
                            data=channel_waveform_data,
                            sample_rate=sr,
                            is_normalized=True,
                        ),
                        sample_rate=self.sample_rate,
                    )
                    for channel_waveform_data in arr
                ]
            )

        self.duration = librosa.get_duration(y=arr, sr=sr)
        return self

    def resample(self, sample_rate: int) -> "Audio":
        """Resample the audio file to the target sample rate.

        Args:
            sample_rate (int): Target sample rate.
        """
        if self.channels is not None:
            self.resample_channel(sample_rate)
        return self

    def resample_channel(self, sample_rate: int) -> "Audio":
        """Resample the audio channel waveform to the target sample rate.

        Args:
            sample_rate (int): Target sample rate.
        """
        if self.channels is None:
            logger.warning(
                f"Audio {self.id} resample failed because it has no channels."
            )
            return self
        for channel in self.channels:
            channel: AudioChannel
            channel.resample(sample_rate)
        return self

    def append_channel(self, channel: AudioChannel | None = None) -> "Audio":
        """Append a channel to the audio file.

        Args:
            channel (AudioChannel): Channel to append.
        """
        if self.channels is None:
            self.channels = AudioChannelList()
        if channel is None:
            channel = AudioChannel()
        self.channels.append(channel)
        return self

    def align_channel_bad(self):
        """Align the bad status of the audio file with its channels."""
        if self.channels is None:
            return
        for channel in self.channels:
            if channel.is_bad:
                self.is_bad = True
                self.bad_reason = channel.bad_reason
                self.bad_component = channel.bad_component
                break

    def align_segment_bad(self):
        """Align the bad status of the audio file with its channels."""
        if self.channels is None:
            return
        for channel in self.channels:
            for span in channel.segments:
                if span.is_bad:
                    self.is_bad = True
                    self.bad_reason = span.bad_reason
                    self.bad_component = span.bad_component
                    break

    def clear(self, clear_text: bool = False) -> None:
        """Clear the waveform, features, and scores of the audio file. should be called after processing. this method is used to save memory.

        Args:
            clear_text
        """
        self.clear_waveform()
        if clear_text:
            self.clear_text()
        return self

    def clear_waveform(self) -> "Audio":
        """clear all waveform data"""
        if self.channels is not None:
            for channel in self.channels:
                channel: AudioChannel
                channel.waveform = None
                if channel.segments is not None:
                    for segment in channel.segments:
                        segment: AudioSpan
                        segment.waveform = None
        return self

    def clear_text(self) -> "Audio":
        """Clear the text of the audio file.

        Returns:
            Audio: Audio object.
        """
        if self.channels is not None:
            for channel in self.channels:
                channel: AudioChannel
                channel.raw_text = None
                channel.tokens = None
                channel.segments = None
                channel.sents = None
        return self

    def clear_bad(self) -> "Audio":
        """Clear the bad status of the audio file.

        Returns:
            Audio: Audio object.
        """
        self.is_bad = None
        self.bad_reason = None
        self.bad_component = None
        return self

    def save(
        self,
        path: str | Path,
        format: str | None = None,
        sample_rate: int | None = None,
    ) -> Path:
        """Save the audio waveform to a file.

        Args:
            path: Output file path.
            format: Audio format (wav, mp3, flac, ogg, etc.).
                If None, inferred from path extension.
            sample_rate: Target sample rate. If None, use original.

        Returns:
            Path to the saved file.
        """
        import soundfile as sf

        path = Path(path)
        fmt = format or path.suffix.lstrip(".")
        if not fmt:
            raise ValueError(
                "Cannot infer format from path, specify format explicitly."
            )

        if not self.channels or not self.channels[0].waveform:
            raise ValueError("Audio has no loaded waveform. Call load() first.")

        if len(self.channels) == 1:
            data = np.asarray(self.channels[0].waveform.data)
            sr = self.channels[0].waveform.sample_rate
        else:
            data = np.stack(
                [np.asarray(ch.waveform.data) for ch in self.channels if ch.waveform],
                axis=-1,
            )
            sr = self.channels[0].waveform.sample_rate

        if sample_rate and sample_rate != sr:
            data = librosa.resample(data, orig_sr=sr, target_sr=sample_rate)
            sr = sample_rate

        if fmt.lower() in ("wav", "flac", "ogg"):
            sf.write(str(path), data, sr)
        else:
            data_int16 = (np.clip(data, -1.0, 1.0) * 32767).astype(np.int16)
            n_channels = 1 if data.ndim == 1 else data.shape[-1]
            seg = AudioSegment(
                data=data_int16.tobytes(),
                sample_width=2,
                frame_rate=sr,
                channels=n_channels,
            )
            seg.export(str(path), format=fmt)

        return path

    def chunk(self, chunk_size_ms: int) -> AudioChunkList:
        """Chunk the audio file into smaller chunks.
        Args:
            chunk_size_ms (int): Size of the chunk in milliseconds.
        Returns:
            AudioList: Chunked audio list.
        """
        if self.channels is None or len(self.channels) == 0:
            raise ValueError("Cannot chunk audio without channels.")
        chunked_audios = AudioChunkList()
        chunk_stride = self.channels[0].waveform.sample_rate / 1000 * chunk_size_ms
        num_chunks = num_chunks = int(np.ceil(self.duration_ms / chunk_size_ms))
        if self.duration_ms > chunk_size_ms:
            for i in range(num_chunks):
                start = int(i * chunk_stride)
                end = int((i + 1) * chunk_stride) if i < num_chunks - 1 else None
                is_start = i == 0
                is_last = i == num_chunks - 1
                chunk_audio = AudioChunk(
                    stream_id=self.id,
                    is_last=is_last,
                    is_start=is_start,
                    waveform=self.channels[0].waveform[start:end],
                )
                chunked_audios.append(chunk_audio)
        else:
            chunk_audio = AudioChunk(
                stream_id=self.id,
                is_last=True,
                waveform=self.channels[0].waveform,
                is_start=True,
            )
            chunked_audios.append(chunk_audio)
        return chunked_audios

    def split_into_spans(self, span_duration_ms: int) -> "AudioSpanList":
        """Split the audio file into smaller spans.

        Args:
            span_duration_ms (int): Duration of the span in milliseconds.

        Returns:
            AudioSpanList: List of spans.
        """
        chunked_audios = AudioSpanList()
        chunk_stride = self.sample_rate / 1000 * span_duration_ms
        num_chunks = int(np.ceil(self.duration_ms / span_duration_ms))
        if self.duration_ms > span_duration_ms:
            for i in range(num_chunks):
                start = int(i * chunk_stride)
                start_ms = i * span_duration_ms
                end = int((i + 1) * chunk_stride) if i < num_chunks - 1 else None
                if end is not None:
                    end_ms = (i + 1) * span_duration_ms
                else:
                    end_ms = self.duration_ms
                is_last = i == num_chunks - 1
                if self[0].waveform is not None:
                    audio_chunk_waveform = self[0].waveform[start:end]
                else:
                    audio_chunk_waveform = None
                chunk_audio = AudioSpan(
                    id=self.id,
                    start_ms=start_ms,
                    end_ms=end_ms,
                    waveform=audio_chunk_waveform,
                    sample_rate=self.sample_rate,
                    duration=span_duration_ms / 1000,
                    is_last=is_last,
                )
                chunked_audios.append(chunk_audio)
        else:
            chunked_audios.append(self)
        return chunked_audios

    def display(self, channel_idx: int = 0):
        """Display the audio file."""
        from IPython.display import Audio as IPAudio

        return IPAudio(
            data=self.channels[channel_idx].waveform.data,
            rate=self.channels[channel_idx].waveform.sample_rate,
        )

    def set_hotwords(self, hotwords: List[str]) -> "Audio":
        """Set the hotwords of the audio file.

        Args:
            hotwords (List[str]): List of hotwords.
        """
        self.hotwords = frozenset(hotwords)
        return self

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        protocol: Literal[
            "protobuf", "pickle", "json", "json-array", "protobuf-array", "pickle-array"
        ] = "protobuf",
        compress: str | None = None,
    ):
        doc = super().from_bytes(data=data, protocol=protocol, compress=compress)
        return doc

    def from_audio_bytes(self, audio_bytes: bytes) -> "Audio":
        """Create an Audio object from audio bytes.

        Args:
            audio_bytes (bytes): Audio bytes.

        Returns:
            Audio: Audio object.
        """
        bytes_io = BytesIO(audio_bytes)
        audio = AudioSegment.from_file(bytes_io)
        bytes_io = BytesIO()
        audio.export(bytes_io, format="wav")
        arr, sr = librosa.load(bytes_io, sr=self.sample_rate, mono=self.mono)
        if len(arr.shape) == 1:
            self.mono = True
            self.channels = AudioChannelList(
                [
                    AudioChannel(
                        id=self.id,
                        waveform=Waveform(data=arr, sample_rate=sr, is_normalized=True),
                    )
                ]
            )
        else:
            self.mono = False
            self.channels = AudioChannelList(
                [
                    AudioChannel(
                        id=self.id,
                        waveform=Waveform(
                            data=channel_waveform_data,
                            sample_rate=sr,
                            is_normalized=True,
                        ),
                        sample_rate=self.sample_rate,
                    )
                    for channel_waveform_data in arr
                ]
            )

        self.duration = librosa.get_duration(y=arr, sr=sr)
        return self

    def from_b64_str(self, b64_str: str) -> "Audio":
        """Create an Audio object from a base64 encoded string.

        Args:
            b64_str (str): Base64 encoded string of the audio file.

        Returns:
            Audio: Audio object.
        """
        import base64

        audio_bytes = base64.b64decode(b64_str.encode("utf-8"))
        return self.from_audio_bytes(audio_bytes)

    @property
    def is_loaded(self):
        if self.channels is None:
            return False
        if len(self.channels) == 0:
            return False
        for channel in self.channels:
            if channel.waveform is None:
                return False
        return True

    @property
    def duration_s(self):
        """Get the duration of the audio file in seconds."""
        if self.duration is not None:
            return self.duration
        if self.channels is None:
            return 0
        if len(self.channels) == 0:
            return 0
        return round(
            len(self.channels[0].waveform.data) / self.channels[0].waveform.sample_rate,
            4,
        )

    @property
    def duration_ms(self):
        """Get the duration of the audio file in milliseconds."""
        return self.duration_s * 1000

    @property
    def audio_format(self) -> str | None:
        if self.url is None:
            return None
        formats = [
            ".mp3",
            ".wav",
            ".amr",
            ".m4a",
            ".flac",
            ".ogg",
            ".opus",
            ".webm",
            ".aac",
            ".wma",
        ]
        for format in formats:
            if format in self.url.lower():
                return format.replace(".", "")
        return None

    def __getitem__(self, idx) -> AudioChannel | None:
        if self.channels is None:
            return None
        return self.channels[idx]

    def __len__(self) -> int:
        if self.channels is None:
            return 0
        return len(self.channels)


class AudioList(DocList[Audio]):
    def load_stream(self, num_threads: int = -1) -> Iterable[Audio]:
        """Load all the audio files in parallel.

        Args:
            num_threads (int, optional): Number of threads to use. Defaults to -1. If -1, use all available cores.
        """
        if len(self) == 0:
            return
        if num_threads == -1:
            num_threads = get_cpu_cores()
        batch_size = max(len(self) // num_threads, 1)
        res = Parallel(
            n_jobs=num_threads,
            prefer="threads",
            batch_size=batch_size,
            return_as="generator_unordered",
            pre_dispatch="4 * n_jobs",
        )(delayed(doc.load)() for doc in self)
        return res

    def load(self, num_workers: int = 2, lazy: bool = False, chunk_size: int = 100):
        """Load all the audio files in parallel.

        Args:
            num_workers (int): Number of workers to use. Defaults to 2. If -1, use all available cores.
            lazy (bool): If True, return a generator that loads audio in chunks
                to limit memory usage. Defaults to False.
            chunk_size (int): Number of audio files to load per chunk when lazy=True.
                Defaults to 100.
        """
        if len(self) == 0:
            return self if not lazy else iter([])
        if num_workers == -1:
            num_workers = get_cpu_cores()
        if lazy:
            return self._lazy_load(num_workers=num_workers, chunk_size=chunk_size)
        batch_size = max(len(self) // num_workers, 1)
        _ = Parallel(n_jobs=num_workers, prefer="threads", batch_size=batch_size)(
            delayed(doc.load)() for doc in self
        )
        return self

    def _lazy_load(
        self, num_workers: int = 2, chunk_size: int = 100
    ) -> Generator[Audio, None, None]:
        """Lazily load audio files in chunks to limit memory usage.

        Each chunk of audio files is loaded in parallel, yielded one by one,
        and cleared before the next chunk is loaded.

        Args:
            num_workers (int): Number of workers to use.
            chunk_size (int): Number of audio files to load per chunk.

        Yields:
            Audio: Loaded audio object. Waveform data is cleared after
                the next chunk begins loading.
        """
        for i in range(0, len(self), chunk_size):
            chunk = self[i : i + chunk_size]
            batch_size = max(len(chunk) // num_workers, 1)
            _ = Parallel(n_jobs=num_workers, prefer="threads", batch_size=batch_size)(
                delayed(doc.load)() for doc in chunk
            )
            yield from chunk
            for doc in chunk:
                doc.clear()

    @classmethod
    def from_urls(
        cls, urls: Union[str, List[str]], load: bool = False, num_workers: int = 2
    ):
        """Create an AudioList from a list of URLs.

        Args:
            urls (List[str]): List of URLs.
            load (bool, optional): Whether to load the audio files. Defaults to False.
            num_workers (int, optional): Number of workers to use. Defaults to -1. If -1, use all available cores.

        Returns:
            AudioList: List of Audio objects.
        """
        if isinstance(urls, str):
            urls = [urls]
        audios = cls([Audio(url=url) for url in urls])
        if load:
            audios = audios.load(num_workers=num_workers)
        return audios

    def resample_channel(self, sample_rate: int, num_workers: int = 2):
        """Resample all the audio files in parallel.

        Args:
            sample_rate (int): Target sample rate.
            num_workers (int, optional): Number of workers to use. Defaults to -1. If -1, use all available cores.
        """
        if num_workers == -1:
            num_workers = get_cpu_cores()
        batch_size = max(len(self) // num_workers, 1)
        _ = Parallel(n_jobs=num_workers, prefer="threads", batch_size=batch_size)(
            delayed(doc.resample_channel)(sample_rate) for doc in self
        )
        return self

    def filter_audio_id(self, ids: List[str], op: str = "$in") -> Optional["AudioList"]:
        """Filter the audio files by their IDs.
        Args:
            ids (List[str]): List of audio IDs.
            op (str, optional): Operator to use. Defaults to "$in". Can be "$in", "$nin"

        Returns:
            Optional[AudioList]: Filtered audio files.
        """
        query = {"id": {op: ids}}
        audios: AudioList = AudioList(filter_docs(self, query=query))
        return audios

    def filter_urls(self, urls: List[str], op: str = "$in") -> Optional["AudioList"]:
        """Filter the audio files by their URLs.
        Args:
            urls (List[str]): List of audio URLs.
            op (str, optional): Operator to use. Defaults to "$in". Can be "$in", "$nin"

        Returns:
            Optional[AudioList]: Filtered audio files.
        """
        query = {"url": {op: urls}}
        audios: AudioList = filter_docs(self, query=query)
        if len(audios) == 0:
            return None
        return audios

    def analysis_timer(self):
        """Print the time spent on processing the audio files."""
        component_spent = {}
        for audio in self:
            for component, spent in audio.spent_time.items():
                if component not in component_spent:
                    component_spent[component] = 0
                component_spent[component] += spent
        total_spent = sum(component_spent.values())
        print("Total spent time: ", round(total_spent, 2))
        for component, spent in component_spent.items():
            print(f"{component}: {spent:.2f}s, {spent / total_spent:.2%}")

    def clear(self) -> None:
        """Clear the waveform, features, and scores of all the audio files. should be called after processing. this method is used to save memory."""
        for audio in self:
            audio.clear()

    def has_bad_audio(self) -> bool:
        """Check if there is bad audio in the list"""
        ids = [audio.id for audio in self if audio.is_bad]
        return len(ids) > 0

    def save_binary(
        self,
        file: str | Path,
        protocol: Literal[
            "protobuf", "pickle", "json", "json-array", "protobuf-array", "pickle-array"
        ] = "pickle",
        compress: Literal["lz4", "bz2", "lzma", "zlib", "gzip"] | None = None,
        show_progress: bool = False,
    ) -> None:
        """Save the audio list to a binary file.

        Args:
            file (str | Path): Path to the binary file.
            protocol (Literal[ &quot;protobuf&quot;, &quot;pickle&quot;, &quot;json&quot;, &quot;json, optional): the protocol to use. Defaults to &quot;pickle&quot;.
            compress (Literal[&quot;lz4&quot;, &quot;bz2&quot;, &quot;lzma&quot;, &quot;zlib&quot;, &quot;gzip&quot;] | None, optional): the compression method to use. Defaults to None.
            show_progress (bool, optional): Whether to show the progress bar. Defaults to False.
        """
        return super().save_binary(file, protocol, compress, show_progress)

    @classmethod
    def load_binary(
        cls,
        file: str | bytes | Path,
        protocol: Literal[
            "protobuf", "pickle", "json", "json-array", "protobuf-array", "pickle-array"
        ] = "pickle",
        compress: Literal["lz4", "bz2", "lzma", "zlib", "gzip"] | None = None,
        show_progress: bool = False,
        streaming: bool = False,
    ) -> DocList[Audio] | Generator[Any, None, None]:
        """Load the audio list from a binary file.

        Args:
            file (str | Path): Path to the binary file.
            show_progress (bool, optional): Whether to show the progress bar. Defaults to False.
            streaming (bool, optional): Whether to stream the data. Defaults to False.
            protocol (Literal[ &quot;protobuf&quot;, &quot;pickle&quot;, &quot;json&quot;, &quot;json, optional): the protocol to use. Defaults to &quot;pickle&quot;.
            compress (Literal[&quot;lz4&quot;, &quot;bz2&quot;, &quot;lzma&quot;, &quot;zlib&quot;, &quot;gzip&quot;] | None, optional): the compression method to use. Defaults to None.

        Returns:
            AudioList: Audio list.
        """
        docs = DocList[Audio].load_binary(
            file=file,
            protocol=protocol,
            compress=compress,
            show_progress=show_progress,
            streaming=streaming,
        )
        if isinstance(docs, Generator):
            return docs
        return AudioList(docs)

    def copy(self) -> "AudioList":
        docs = [audio.model_copy(deep=True) for audio in self]
        return AudioList(docs)

    def clear(self, clear_text: bool = False) -> "AudioList":  # noqa
        for audio in self:
            audio: Audio
            audio.clear(clear_text=clear_text)
        return self

    def clear_waveform(self) -> "AudioList":
        for audio in self:
            audio: Audio
            audio.clear_waveform()
        return self

    def clear_text(self) -> "AudioList":
        for audio in self:
            audio: Audio
            audio.clear_text()
        return self

    def save_audios(
        self,
        output_dir: str | Path,
        format: str = "wav",
        sample_rate: int | None = None,
        load: bool = True,
        num_workers: int = 2,
    ) -> Path:
        """Save all audio waveforms to individual files in a directory.

        Args:
            output_dir: Output directory path.
            format: Audio format (wav, mp3, flac, ogg, etc.).
            sample_rate: Target sample rate. If None, use original.
            load: If True, lazily load each audio from url before saving
                and clear after, keeping only one audio in memory at a time.
            num_workers: Number of parallel workers (only used when load=False).

        Returns:
            Path to the output directory.
        """
        from concurrent.futures import ThreadPoolExecutor

        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        width = len(str(len(self) - 1))

        def _filename(audio: Audio, index: int) -> str:
            if audio.url:
                stem = Path(str(audio.url).split("?")[0]).stem
            else:
                stem = str(index).zfill(width)
            return f"{stem}.{format}"

        def _save_one(idx_audio):
            idx, audio = idx_audio
            path = output_dir / _filename(audio, idx)
            audio.save(path, format=format, sample_rate=sample_rate)

        if load:
            for i, audio in enumerate(self):
                audio.load()
                path = output_dir / _filename(audio, i)
                audio.save(path, format=format, sample_rate=sample_rate)
                audio.clear()
        else:
            if num_workers <= 1:
                for i, audio in enumerate(self):
                    _save_one((i, audio))
            else:
                with ThreadPoolExecutor(max_workers=num_workers) as pool:
                    list(pool.map(_save_one, enumerate(self)))

        return output_dir

    def group_by_hotwords(self) -> Dict[FrozenSet[str], "AudioList"]:
        """Group the audio files by their hotwords. used for batch processing."""
        groups = {}
        for audio in self:
            hotwords = audio.hotwords
            if hotwords is None:
                hotwords = frozenset()
            if hotwords not in groups:
                groups[hotwords] = AudioList()
            groups[hotwords].append(audio)
        return groups

    def shuffle(self) -> "AudioList":
        """Shuffle the audio files."""
        import random

        random.shuffle(self)
        return self

    def _serialize_audio(self, audio: Audio) -> Dict[str, Any]:
        """Serialize an Audio object to a litdata-compatible flat dict.

        Non-destructive: original Audio objects are not modified.
        Tensor data is extracted into a pickled bytes blob.
        Scalar/text fields go through model_dump_json (waveform fields are
        already exclude=True; span feats/scores are temporarily nullified
        then restored).
        """
        import pickle

        tensors: Dict[str, np.ndarray] = {}
        meta: Dict[str, Any] = {}
        _to_restore: list[tuple[Any, str, Any]] = []

        def _extract_waveform(prefix: str, w: Waveform):
            meta[prefix] = {
                "sample_rate": w.sample_rate,
                "is_normalized": w.is_normalized,
            }
            if w.data is not None:
                tensors[f"{prefix}.data"] = w.data
            if w.feats is not None:
                tensors[f"{prefix}.feats"] = np.asarray(w.feats)
            if w.scores is not None:
                tensors[f"{prefix}.scores"] = np.asarray(w.scores)

        def _extract_spans(prefix: str, spans: AudioSpanList | None):
            if not spans:
                return
            for j, span in enumerate(spans):
                if span.waveform is not None:
                    _extract_waveform(f"{prefix}.{j}.w", span.waveform)
                if span.feats is not None:
                    tensors[f"{prefix}.{j}.feats"] = np.asarray(span.feats)
                    _to_restore.append((span, "feats", span.feats))
                    span.feats = None
                if span.scores is not None:
                    tensors[f"{prefix}.{j}.scores"] = np.asarray(span.scores)
                    _to_restore.append((span, "scores", span.scores))
                    span.scores = None
                if span.tokens:
                    for k, tok in enumerate(span.tokens):
                        if tok.waveform is not None:
                            _extract_waveform(f"{prefix}.{j}.t.{k}", tok.waveform)

        if audio.channels:
            for i, ch in enumerate(audio.channels):
                p = str(i)
                if ch.waveform is not None:
                    _extract_waveform(p, ch.waveform)
                _extract_spans(f"{p}.seg", ch.segments)
                _extract_spans(f"{p}.sent", ch.sents)
                if ch.tokens:
                    for k, tok in enumerate(ch.tokens):
                        if tok.waveform is not None:
                            _extract_waveform(f"{p}.tok.{k}", tok.waveform)

        result = {
            "audio_json": audio.model_dump_json(),
            "audio_blob": pickle.dumps({"tensors": tensors, "meta": meta}),
        }

        for obj, attr, val in _to_restore:
            setattr(obj, attr, val)

        return result

    @staticmethod
    def _deserialize_audio(item: Dict[str, Any]) -> "Audio":
        """Deserialize a litdata item back into an Audio object."""
        import pickle

        audio = Audio.model_validate_json(item["audio_json"])
        blob = pickle.loads(item["audio_blob"])
        tensors: Dict[str, np.ndarray] = blob.get("tensors", {})
        meta: Dict[str, Any] = blob.get("meta", {})

        def _restore_waveform(prefix: str) -> Waveform | None:
            data = tensors.get(f"{prefix}.data")
            feats = tensors.get(f"{prefix}.feats")
            scores = tensors.get(f"{prefix}.scores")
            if data is None and feats is None and scores is None:
                return None
            wm = meta.get(prefix, {})
            return Waveform(
                data=data,
                sample_rate=wm.get("sample_rate"),
                is_normalized=wm.get("is_normalized"),
                feats=feats,
                scores=scores,
            )

        def _restore_spans(prefix: str, spans: AudioSpanList | None):
            if not spans:
                return
            for j, span in enumerate(spans):
                w = _restore_waveform(f"{prefix}.{j}.w")
                if w is not None:
                    span.waveform = w
                span_feats = tensors.get(f"{prefix}.{j}.feats")
                if span_feats is not None:
                    span.feats = span_feats
                span_scores = tensors.get(f"{prefix}.{j}.scores")
                if span_scores is not None:
                    span.scores = span_scores
                if span.tokens:
                    for k, tok in enumerate(span.tokens):
                        tw = _restore_waveform(f"{prefix}.{j}.t.{k}")
                        if tw is not None:
                            tok.waveform = tw

        if audio.channels:
            for i, ch in enumerate(audio.channels):
                p = str(i)
                w = _restore_waveform(p)
                if w is not None:
                    ch.waveform = w
                _restore_spans(f"{p}.seg", ch.segments)
                _restore_spans(f"{p}.sent", ch.sents)
                if ch.tokens:
                    for k, tok in enumerate(ch.tokens):
                        tw = _restore_waveform(f"{p}.tok.{k}")
                        if tw is not None:
                            tok.waveform = tw
        return audio

    def to_dataset(
        self,
        dataset_dir: str | Path,
        num_workers: int = 2,
        mode: Literal["overwrite", "append"] = "overwrite",
        chunk_bytes: str = "60MB",
        load: bool = False,
    ):
        """Convert the audio list to a litdata format streamdataset.

        Serializes each Audio object completely, preserving all channels,
        segments, tokens, hotwords, and metadata.

        Args:
            dataset_dir (str | Path): Path to the dataset directory.
            num_workers (int): Number of workers to use (ignored when load=True).
            mode (Literal["overwrite", "append"]): Whether to overwrite or append.
            chunk_bytes (str): Chunk size. Defaults to "60MB".
            load (bool): If True, load each audio from url before serialization
                and clear after. Uses streaming write so only one audio's
                waveform is in memory at a time.
        """
        import shutil

        dataset_dir = str(dataset_dir)

        if load:
            from litdata.streaming.cache import Cache

            if mode == "overwrite" and Path(dataset_dir).exists():
                shutil.rmtree(dataset_dir)
            Path(dataset_dir).mkdir(parents=True, exist_ok=True)

            cache = Cache(input_dir=dataset_dir, chunk_bytes=chunk_bytes)
            for i, audio in enumerate(self):
                audio.load()
                cache[i] = self._serialize_audio(audio)
                audio.clear()
            cache.done()
            cache.merge()
        else:
            from litdata import optimize

            def get_item(index: int) -> Dict[str, Any]:
                return self._serialize_audio(self[index])

            optimize(
                fn=get_item,
                inputs=list(range(len(self))),
                output_dir=dataset_dir,
                num_workers=num_workers,
                chunk_bytes=chunk_bytes,
                mode=mode,
            )

    @classmethod
    def from_dataset(cls, dataset_dir: str | Path) -> "AudioList":
        """Load litdata streamdataset to AudioList."""
        from litdata import StreamingDataset

        ds = StreamingDataset(input_dir=dataset_dir)
        audios = AudioList()
        for item in ds:
            audios.append(cls._deserialize_audio(item))
        return audios

    @property
    def duration_s(self):
        """Get the total duration of all the audio files in seconds."""
        durations = []
        for i in range(len(self)):
            audio: Audio = self[i]
            if audio.duration is not None:
                durations.append(audio.duration)
            else:
                durations.append(audio.duration_s)
        return round(sum(durations), 4)

    @property
    def duration_ms(self):
        """Get the total duration of all the audio files in milliseconds."""
        durations = []
        for audio in self:
            if audio.duration_ms is not None:
                durations.append(audio.duration_ms)
        return round(sum(durations), 4)

    @property
    def channels(self):
        """Get all the channels of the audio files."""
        channels = []
        for audio in self:
            if audio.channels is not None:
                channels.extend(audio.channels)
        return channels

    @property
    def bad_audios(self) -> "AudioList":
        """Get the bad audios in the list"""
        return AudioList([audio for audio in self if audio.is_bad])

    @property
    def good_audios(self) -> "AudioList":
        """Get the good audios in the list"""
        return AudioList([audio for audio in self if not audio.is_bad])

    def __sub__(self, other: "AudioList") -> "AudioList":
        """Get the difference between two AudioList objects."""
        other_ids = [audio.id for audio in other]
        return self.filter_audio_id(other_ids, op="$nin")

    def __str__(self) -> str:
        num_bad_audios = [audio.is_bad for audio in self].count(True)
        return f"AudioList with {len(self)} audios, {num_bad_audios} bad audios."

    def __repr__(self) -> str:
        return self.__str__()


@lru_cache
def get_cpu_cores():
    return os.cpu_count()


class AudioChunk(BaseDoc):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="ignore")

    id: ID = None
    stream_id: ID
    waveform: Waveform
    start_ms: int | None = None
    end_ms: int | None = None
    tokens: AudioTokenList | None = None
    is_start: bool = False
    is_last: bool = False
    vad_state: Literal["segment_start", "segment_end", "segment_mid"] | None = None

    @property
    def text(self) -> str:
        """Get the text of the audio chunk."""
        if self.tokens is None:
            return ""
        text = ""
        for token in self.tokens:
            text += token.text
        return text

    @property
    def is_final(self) -> bool:
        """Get the is_final of the audio chunk."""
        return self.is_last

    @property
    def duration_ms(self) -> float:
        """Get the duration of the audio chunk in milliseconds."""
        return self.waveform.duration_ms

    @property
    def duration_s(self) -> float:
        """Get the duration of the audio chunk in seconds."""
        return self.waveform.duration_s

    def clear(self) -> "AudioChunk":
        self.waveform = None
        self.tokens = None
        return self

    def display(self):
        """Display the audio chunk."""
        from IPython.display import Audio as IPAudio

        return IPAudio(data=self.waveform.data, rate=self.waveform.sample_rate)


class AudioChunkList(DocList[AudioChunk]):
    def concat(self) -> Audio:
        """Concatenate all the audio chunks into a single audio file."""
        if len(self) == 0:
            return Audio()
        waveform = Waveform.concatenate([chunk.waveform for chunk in self])
        audio = Audio(sample_rate=waveform.sample_rate)
        audio.append_channel(AudioChannel(waveform=waveform))
        return audio

    def concat_to_chunk(self) -> AudioChunk:
        """Concatenate all the audio chunks into a single audio chunk.

        note: this method will set start_ms, end_ms, and stream_id of the audio chunk.
        if there is only one audio chunk, return the audio chunk directly.
        if there are multiple audio chunks, return a new audio chunk with concatenated waveform.
        if there are no audio chunks, return an empty audio chunk.
        """
        if len(self) == 0:
            return AudioChunk()
        if len(self) == 1:
            return self[0]
        stream_id = self[0].stream_id
        start_ms = self[0].start_ms
        end_ms = self[-1].end_ms
        waveform = Waveform.concatenate([chunk.waveform for chunk in self])
        audio_chunk = AudioChunk(
            waveform=waveform, start_ms=start_ms, end_ms=end_ms, stream_id=stream_id
        )
        return audio_chunk

    def concat_to_audio(self) -> Audio:
        """Concatenate all the audio chunks into a single audio file.
        note: this method will set the is_last property of the audio chunk.
        """
        if len(self) == 0:
            return Audio()
        if len(self) == 1:
            self[0].is_last = True
            return self[0].concat()
        waveform = Waveform.concatenate([chunk.waveform for chunk in self])
        audio = Audio(sample_rate=waveform.sample_rate)
        audio.append_channel(AudioChannel(waveform=waveform))
        return audio



import pytest

from fasr.utils.metric import compute_char_error_rate, match_token, streaming_delay


class TestComputeCharErrorRate:
    def test_identical_strings(self):
        assert compute_char_error_rate("你好世界", "你好世界") == 0.0

    def test_completely_different(self):
        cer = compute_char_error_rate("abc", "xyz")
        assert cer == 1.0

    def test_partial_match(self):
        cer = compute_char_error_rate("你好世界", "你好地球")
        assert 0 < cer < 1

    def test_empty_both(self):
        cer = compute_char_error_rate("", "")
        assert cer == 1.0

    def test_empty_target(self):
        cer = compute_char_error_rate("hello", "")
        assert cer == 0.0

    def test_empty_source(self):
        cer = compute_char_error_rate("", "hello")
        assert cer > 0

    def test_ignore_space(self):
        cer = compute_char_error_rate("你 好", "你好", ignore_space=True)
        assert cer == 0.0

    def test_keep_space(self):
        cer_ignore = compute_char_error_rate("你 好", "你好", ignore_space=True)
        cer_keep = compute_char_error_rate("你 好", "你好", ignore_space=False)
        assert cer_keep >= cer_ignore

    def test_ignore_punctuation(self):
        cer = compute_char_error_rate("你好，世界！", "你好世界", ignore_punc=True)
        assert cer == 0.0

    def test_keep_punctuation(self):
        cer = compute_char_error_rate("你好，世界！", "你好世界", ignore_punc=False)
        assert cer > 0

    def test_insertion(self):
        cer = compute_char_error_rate("你好啊世界", "你好世界")
        assert cer > 0

    def test_deletion(self):
        cer = compute_char_error_rate("你好", "你好世界")
        assert cer > 0

    def test_substitution(self):
        cer = compute_char_error_rate("你坏世界", "你好世界")
        assert cer > 0


class TestMatchToken:
    def test_identical(self):
        matched = match_token("abc", "abc")
        assert len(matched) == 3
        assert matched == [(0, 0), (1, 1), (2, 2)]

    def test_insertion(self):
        matched = match_token("axbc", "abc")
        assert len(matched) == 3

    def test_deletion(self):
        matched = match_token("ac", "abc")
        assert len(matched) == 2

    def test_substitution(self):
        matched = match_token("axc", "abc")
        assert len(matched) == 2

    def test_case_insensitive(self):
        matched = match_token("ABC", "abc")
        assert len(matched) == 3

    def test_empty_strings(self):
        matched = match_token("", "")
        assert matched == []


class TestStreamingDelay:
    def test_basic(self):
        t_send = [
            {"token": "a", "timestamp": 0.1},
            {"token": "b", "timestamp": 0.2},
            {"token": "c", "timestamp": 0.3},
        ]
        t_response = [
            {"token": "a", "timestamp": 0.3},
            {"token": "b", "timestamp": 0.4},
            {"token": "c", "timestamp": 0.5},
        ]
        delay = streaming_delay(t_send, t_response)
        assert isinstance(delay, float)
        assert abs(delay - 0.2) < 0.01

    def test_return_results(self):
        t_send = [
            {"token": "a", "timestamp": 0.1},
            {"token": "b", "timestamp": 0.2},
        ]
        t_response = [
            {"token": "a", "timestamp": 0.3},
            {"token": "b", "timestamp": 0.4},
        ]
        results = streaming_delay(t_send, t_response, return_results=True)
        assert isinstance(results, list)
        assert len(results) == 2
        assert results[0]["token"] == "a"
        assert abs(results[0]["delay"] - 0.2) < 0.01

    def test_partial_match(self):
        t_send = [
            {"token": "a", "timestamp": 0.1},
            {"token": "b", "timestamp": 0.2},
            {"token": "c", "timestamp": 0.3},
        ]
        t_response = [
            {"token": "a", "timestamp": 0.3},
            {"token": "c", "timestamp": 0.5},
        ]
        results = streaming_delay(t_send, t_response, return_results=True)
        assert len(results) == 2



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

import os
import sys
import warnings

warnings.filterwarnings("ignore")
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"  # 0=ALL, 1=INFO, 2=WARNING, 3=ERROR
os.environ["TF_ENABLE_ONEDNN_OPTS"] = "0"

__version__ = "0.0.1"

_CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

try:
    from fireredasr2s.fireredasr2.asr import FireRedAsr2, FireRedAsr2Config
except ImportError:
    if _CURRENT_DIR not in sys.path:
        sys.path.insert(0, _CURRENT_DIR)
    from .asr import FireRedAsr2, FireRedAsr2Config


# API
__all__ = [
    "__version__",
    "FireRedAsr2",
    "FireRedAsr2Config",
]



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Yan Jia)

import torch
import torch.nn as nn
import torch.nn.functional as F


class ConformerEncoder(nn.Module):
    def __init__(
        self,
        idim,
        n_layers,
        n_head,
        d_model,
        residual_dropout=0.1,
        dropout_rate=0.1,
        kernel_size=33,
        pe_maxlen=5000,
    ):
        super().__init__()
        self.odim = d_model

        self.input_preprocessor = Conv2dSubsampling(idim, d_model)
        self.positional_encoding = RelPositionalEncoding(d_model)
        self.dropout = nn.Dropout(residual_dropout)

        self.layer_stack = nn.ModuleList()
        for l in range(n_layers):
            block = RelPosEmbConformerBlock(
                d_model, n_head, residual_dropout, dropout_rate, kernel_size
            )
            self.layer_stack.append(block)

    def forward(self, padded_input, input_lengths, pad=True):
        if pad:
            padded_input = F.pad(
                padded_input,
                (0, 0, 0, self.input_preprocessor.context - 1),
                "constant",
                0.0,
            )
        src_mask = self.padding_position_is_0(padded_input, input_lengths)

        embed_output, input_lengths, src_mask = self.input_preprocessor(
            padded_input, src_mask
        )
        enc_output = self.dropout(embed_output)

        pos_emb = self.dropout(self.positional_encoding(embed_output))

        enc_outputs = []
        for enc_layer in self.layer_stack:
            enc_output = enc_layer(
                enc_output, pos_emb, slf_attn_mask=src_mask, pad_mask=src_mask
            )
            enc_outputs.append(enc_output)

        return enc_output, input_lengths, src_mask

    def padding_position_is_0(self, padded_input, input_lengths):
        N, T = padded_input.size()[:2]
        mask = torch.ones((N, T)).to(padded_input.device)
        for i in range(N):
            mask[i, input_lengths[i] :] = 0
        mask = mask.unsqueeze(dim=1)
        return mask.to(torch.uint8)


class RelPosEmbConformerBlock(nn.Module):
    def __init__(
        self,
        d_model,
        n_head,
        residual_dropout=0.1,
        dropout_rate=0.1,
        kernel_size=33,
    ):
        super().__init__()
        self.ffn1 = ConformerFeedForward(d_model, dropout_rate)
        self.mhsa = RelPosMultiHeadAttention(n_head, d_model, residual_dropout)
        self.conv = ConformerConvolution(d_model, kernel_size, dropout_rate)
        self.ffn2 = ConformerFeedForward(d_model, dropout_rate)
        self.layer_norm = nn.LayerNorm(d_model)

    def forward(self, x, pos_emb, slf_attn_mask=None, pad_mask=None):
        out = 0.5 * x + 0.5 * self.ffn1(x)
        out = self.mhsa(out, out, out, pos_emb, mask=slf_attn_mask)[0]
        out = self.conv(out, pad_mask)
        out = 0.5 * out + 0.5 * self.ffn2(out)
        out = self.layer_norm(out)
        return out


class Swish(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)


class Conv2dSubsampling(nn.Module):
    def __init__(self, idim, d_model, out_channels=32):
        super().__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(1, out_channels, 3, 2),
            nn.ReLU(),
            nn.Conv2d(out_channels, out_channels, 3, 2),
            nn.ReLU(),
        )
        subsample_idim = ((idim - 1) // 2 - 1) // 2
        self.out = nn.Linear(out_channels * subsample_idim, d_model)

        self.subsampling = 4
        left_context = right_context = 3
        self.context = left_context + 1 + right_context  # 7

    def forward(self, x, x_mask):
        x = x.unsqueeze(1)
        x = self.conv(x)
        N, C, T, D = x.size()
        x = self.out(x.transpose(1, 2).contiguous().view(N, T, C * D))
        mask = x_mask[:, :, :-2:2][:, :, :-2:2]
        input_lengths = mask[:, -1, :].sum(dim=-1)
        return x, input_lengths, mask


class RelPositionalEncoding(torch.nn.Module):
    def __init__(self, d_model, max_len=5000):
        super().__init__()
        pe_positive = torch.zeros(max_len, d_model, requires_grad=False)
        pe_negative = torch.zeros(max_len, d_model, requires_grad=False)
        position = torch.arange(0, max_len).unsqueeze(1).float()
        div_term = torch.exp(
            torch.arange(0, d_model, 2).float()
            * -(torch.log(torch.tensor(10000.0)).item() / d_model)
        )
        pe_positive[:, 0::2] = torch.sin(position * div_term)
        pe_positive[:, 1::2] = torch.cos(position * div_term)
        pe_negative[:, 0::2] = torch.sin(-1 * position * div_term)
        pe_negative[:, 1::2] = torch.cos(-1 * position * div_term)

        pe_positive = torch.flip(pe_positive, [0]).unsqueeze(0)
        pe_negative = pe_negative[1:].unsqueeze(0)
        pe = torch.cat([pe_positive, pe_negative], dim=1)
        self.register_buffer("pe", pe)

    def forward(self, x):
        Tmax, T = self.pe.size(1), x.size(1)
        pos_emb = self.pe[:, Tmax // 2 - T + 1 : Tmax // 2 + T].clone().detach()
        return pos_emb


class ConformerFeedForward(nn.Module):
    def __init__(self, d_model, dropout_rate=0.1):
        super().__init__()
        pre_layer_norm = nn.LayerNorm(d_model)
        linear_expand = nn.Linear(d_model, d_model * 4)
        nonlinear = Swish()
        dropout_pre = nn.Dropout(dropout_rate)
        linear_project = nn.Linear(d_model * 4, d_model)
        dropout_post = nn.Dropout(dropout_rate)
        self.net = nn.Sequential(
            pre_layer_norm,
            linear_expand,
            nonlinear,
            dropout_pre,
            linear_project,
            dropout_post,
        )

    def forward(self, x):
        residual = x
        output = self.net(x)
        output = output + residual
        return output


class ConformerConvolution(nn.Module):
    def __init__(self, d_model, kernel_size=33, dropout_rate=0.1):
        super().__init__()
        assert kernel_size % 2 == 1
        self.pre_layer_norm = nn.LayerNorm(d_model)
        self.pointwise_conv1 = nn.Conv1d(
            d_model, d_model * 4, kernel_size=1, bias=False
        )
        self.glu = F.glu
        self.padding = (kernel_size - 1) // 2
        self.depthwise_conv = nn.Conv1d(
            d_model * 2,
            d_model * 2,
            kernel_size,
            stride=1,
            padding=self.padding,
            groups=d_model * 2,
            bias=False,
        )
        self.batch_norm = nn.LayerNorm(d_model * 2)
        self.swish = Swish()
        self.pointwise_conv2 = nn.Conv1d(
            d_model * 2, d_model, kernel_size=1, bias=False
        )
        self.dropout = nn.Dropout(dropout_rate)

    def forward(self, x, mask=None):
        residual = x
        out = self.pre_layer_norm(x)
        out = out.transpose(1, 2)
        if mask is not None:
            out.masked_fill_(mask.ne(1), 0.0)
        out = self.pointwise_conv1(out)
        out = F.glu(out, dim=1)
        out = self.depthwise_conv(out)

        out = out.transpose(1, 2)
        out = self.swish(self.batch_norm(out))
        out = out.transpose(1, 2)

        out = self.dropout(self.pointwise_conv2(out))
        if mask is not None:
            out.masked_fill_(mask.ne(1), 0.0)
        out = out.transpose(1, 2)
        return out + residual


class EncoderMultiHeadAttention(nn.Module):
    def __init__(self, n_head, d_model, residual_dropout=0.1):
        super().__init__()
        assert d_model % n_head == 0
        self.n_head = n_head
        self.d_k = d_model // n_head
        self.d_v = self.d_k

        self.w_qs = nn.Linear(d_model, n_head * self.d_k, bias=False)
        self.w_ks = nn.Linear(d_model, n_head * self.d_k, bias=False)
        self.w_vs = nn.Linear(d_model, n_head * self.d_v, bias=False)

        self.layer_norm_q = nn.LayerNorm(d_model)
        self.layer_norm_k = nn.LayerNorm(d_model)
        self.layer_norm_v = nn.LayerNorm(d_model)

        self.attention = ScaledDotProductAttention(temperature=self.d_k**0.5)
        self.fc = nn.Linear(n_head * self.d_v, d_model, bias=False)
        self.dropout = nn.Dropout(residual_dropout)

    def forward(self, q, k, v, mask=None):
        sz_b, len_q = q.size(0), q.size(1)

        residual = q
        q, k, v = self.forward_qkv(q, k, v)

        output, attn = self.attention(q, k, v, mask=mask)

        output = self.forward_output(output, residual, sz_b, len_q)
        return output, attn

    def forward_qkv(self, q, k, v):
        d_k, d_v, n_head = self.d_k, self.d_v, self.n_head
        sz_b, len_q, len_k, len_v = (
            q.size(0),
            q.size(1),
            k.size(1),
            v.size(1),
        )

        q = self.layer_norm_q(q)
        k = self.layer_norm_k(k)
        v = self.layer_norm_v(v)

        q = self.w_qs(q).view(sz_b, len_q, n_head, d_k)
        k = self.w_ks(k).view(sz_b, len_k, n_head, d_k)
        v = self.w_vs(v).view(sz_b, len_v, n_head, d_v)
        q = q.transpose(1, 2)
        k = k.transpose(1, 2)
        v = v.transpose(1, 2)
        return q, k, v

    def forward_output(self, output, residual, sz_b, len_q):
        output = output.transpose(1, 2).contiguous().view(sz_b, len_q, -1)
        fc_out = self.fc(output)
        output = self.dropout(fc_out)
        output = output + residual
        return output


class ScaledDotProductAttention(nn.Module):
    def __init__(self, temperature):
        super().__init__()
        self.temperature = temperature
        self.dropout = nn.Dropout(0.0)
        self.INF = float("inf")

    def forward(self, q, k, v, mask=None):
        attn = torch.matmul(q, k.transpose(2, 3)) / self.temperature
        output, attn = self.forward_attention(attn, v, mask)
        return output, attn

    def forward_attention(self, attn, v, mask=None):
        if mask is not None:
            mask = mask.unsqueeze(1)
            mask = mask.eq(0)
            attn = attn.masked_fill(mask, -self.INF)
            attn = torch.softmax(attn, dim=-1).masked_fill(mask, 0.0)
        else:
            attn = torch.softmax(attn, dim=-1)

        d_attn = self.dropout(attn)
        output = torch.matmul(d_attn, v)

        return output, attn


class RelPosMultiHeadAttention(EncoderMultiHeadAttention):
    def __init__(self, n_head, d_model, residual_dropout=0.1):
        super().__init__(n_head, d_model, residual_dropout)
        d_k = d_model // n_head
        self.scale = 1.0 / (d_k**0.5)
        self.linear_pos = nn.Linear(d_model, n_head * d_k, bias=False)
        self.pos_bias_u = nn.Parameter(torch.FloatTensor(n_head, d_k))
        self.pos_bias_v = nn.Parameter(torch.FloatTensor(n_head, d_k))
        torch.nn.init.xavier_uniform_(self.pos_bias_u)
        torch.nn.init.xavier_uniform_(self.pos_bias_v)

    def _rel_shift(self, x):
        N, H, T1, T2 = x.size()
        zero_pad = torch.zeros((N, H, T1, 1), device=x.device, dtype=x.dtype)
        x_padded = torch.cat([zero_pad, x], dim=-1)

        x_padded = x_padded.view(N, H, T2 + 1, T1)
        x = x_padded[:, :, 1:].view_as(x)
        x = x[:, :, :, : x.size(-1) // 2 + 1]
        return x

    def forward(self, q, k, v, pos_emb, mask=None):
        sz_b, len_q = q.size(0), q.size(1)

        residual = q
        q, k, v = self.forward_qkv(q, k, v)

        q = q.transpose(1, 2)
        n_batch_pos = pos_emb.size(0)
        p = self.linear_pos(pos_emb).view(n_batch_pos, -1, self.n_head, self.d_k)
        p = p.transpose(1, 2)

        q_with_bias_u = (q + self.pos_bias_u).transpose(1, 2)
        q_with_bias_v = (q + self.pos_bias_v).transpose(1, 2)

        matrix_ac = torch.matmul(q_with_bias_u, k.transpose(-2, -1))

        matrix_bd = torch.matmul(q_with_bias_v, p.transpose(-2, -1))
        matrix_bd = self._rel_shift(matrix_bd)

        attn_scores = matrix_ac + matrix_bd
        attn_scores.mul_(self.scale)

        output, attn = self.attention.forward_attention(attn_scores, v, mask=mask)

        output = self.forward_output(output, residual, sz_b, len_q)
        return output, attn



import numpy as np
import pytest

from fasr.data.waveform import Waveform


class TestWaveformBasic:
    def test_create(self, sine_wave, sample_rate):
        wf = Waveform(data=sine_wave, sample_rate=sample_rate)
        assert wf.data is not None
        assert wf.sample_rate == sample_rate
        assert len(wf) == len(sine_wave)

    def test_duration(self, waveform, sample_rate):
        assert abs(waveform.duration - 1.0) < 0.01
        assert abs(waveform.duration_s - 1.0) < 0.01
        assert abs(waveform.duration_ms - 1000.0) < 10

    def test_ndim(self, waveform):
        assert waveform.ndim == 1

    def test_len(self, waveform, sample_rate):
        assert len(waveform) == sample_rate


class TestWaveformSlice:
    def test_getitem_int(self, waveform):
        sliced = waveform[0:100]
        assert isinstance(sliced, Waveform)
        assert len(sliced) == 100
        assert sliced.sample_rate == waveform.sample_rate

    def test_getitem_slice(self, waveform, sample_rate):
        half = waveform[: sample_rate // 2]
        assert len(half) == sample_rate // 2

    def test_select_by_ms(self, waveform, sample_rate):
        selected = waveform.select_by_ms(0, 500)
        expected_len = int(500 * sample_rate / 1000)
        assert len(selected) == expected_len
        assert selected.sample_rate == waveform.sample_rate

    def test_select_by_ms_none_start(self, waveform, sample_rate):
        selected = waveform.select_by_ms(None, 500)
        expected_len = int(500 * sample_rate / 1000)
        assert len(selected) == expected_len

    def test_select_by_ms_none_end(self, waveform, sample_rate):
        selected = waveform.select_by_ms(200, None)
        start_idx = int(200 * sample_rate / 1000)
        expected_len = sample_rate - start_idx
        assert len(selected) == expected_len


class TestWaveformResample:
    def test_resample_down(self, waveform):
        resampled = waveform.resample(8000)
        assert resampled.sample_rate == 8000
        assert abs(resampled.duration - 1.0) < 0.05

    def test_resample_up(self, sample_rate):
        data = np.zeros(8000, dtype=np.float32)
        wf = Waveform(data=data, sample_rate=8000)
        resampled = wf.resample(sample_rate)
        assert resampled.sample_rate == sample_rate


class TestWaveformConcat:
    def test_concatenate(self, waveform, short_waveform):
        combined = Waveform.concatenate([waveform, short_waveform])
        assert len(combined) == len(waveform) + len(short_waveform)
        assert combined.sample_rate == waveform.sample_rate

    def test_append(self, sample_rate):
        wf1 = Waveform(data=np.ones(100, dtype=np.float32), sample_rate=sample_rate)
        wf2 = Waveform(data=np.zeros(50, dtype=np.float32), sample_rate=sample_rate)
        wf1.append(wf2)
        assert len(wf1) == 150

    def test_append_different_sr_raises(self):
        wf1 = Waveform(data=np.ones(100, dtype=np.float32), sample_rate=16000)
        wf2 = Waveform(data=np.ones(100, dtype=np.float32), sample_rate=8000)
        with pytest.raises(AssertionError):
            wf1.append(wf2)


class TestWaveformTransform:
    def test_flatten(self, sample_rate):
        data = np.ones((2, 100), dtype=np.float32)
        wf = Waveform(data=data, sample_rate=sample_rate)
        wf.flatten()
        assert wf.ndim == 1
        assert len(wf) == 200

    def test_as_1d_clears_feats(self, sample_rate):
        data = np.ones((2, 100), dtype=np.float32)
        feats = np.ones((10, 80), dtype=np.float32)
        wf = Waveform(data=data, sample_rate=sample_rate, feats=feats, scores=feats)
        wf.as_1d(clear=True)
        assert wf.ndim == 1
        assert wf.feats is None
        assert wf.scores is None

    def test_as_1d_keeps_feats(self, sample_rate):
        data = np.ones((2, 100), dtype=np.float32)
        feats = np.ones((10, 80), dtype=np.float32)
        wf = Waveform(data=data, sample_rate=sample_rate, feats=feats)
        wf.as_1d(clear=False)
        assert wf.ndim == 1
        assert wf.feats is not None
